#include <tclap/CmdLine.h>
#include <common/GlobalData.h>
#include <measure>
#include <string>
#include <exception/AlibException.h>
#include <factory/XmlDataFactory.hpp>
#include <sax/SaxParseInterface.h>
#include <sax/ParserException.h>
#include <graph/isomorphism/Isomorphism.h>
#include <graph/isomorphism/HopcroftWong.h>
#include <graph/embedding/HopcroftTarjan.h>
#include <graph/generate/RandomGraphFactory.h>

int main(int argc, char** argv) {
	try {
		TCLAP::CmdLine cmd("Graph benchmark binary", ' ', "0.01");

		TCLAP::ValueArg<int> size(	"",	"size",	"Size of graph",		false,	10, "integer");
		cmd.add( size );

		TCLAP::SwitchArg onlyhopcroft(		"",	"onlyhopcroft",	"Use only HopcroftWong algorithm",			false);
		cmd.add( onlyhopcroft );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		common::GlobalData::measure = true;
		if(verbose.isSet())
			common::GlobalData::verbose = true;

		graph::UndirectedGraph ug1 = graph::generate::RandomGraphFactory::generateUndirectedTriConnectedPlanarGraph( size.getValue() );
		ug1.setEmbedding(graph::embedding::HopcroftTarjan::hopcrofttarjan(ug1));
		graph::UndirectedGraph ug2 = graph::generate::RandomGraphFactory::generateUndirectedIsomorphicGraph(ug1);
		ug2.setEmbedding(graph::embedding::HopcroftTarjan::hopcrofttarjan(ug2));

		measurements::start("HopcroftWong", measurements::Type::OVERALL);
		if (!graph::isomorphism::HopcroftWong::hopcroftwong(ug1, ug2))
			std::cout << "FAIL: hopcroftwong returned false!" << std::endl;
		measurements::end();
		std::clog << measurements::results() << std::endl;

		if (!onlyhopcroft.isSet()) {
			measurements::start("RecursiveByDefinition", measurements::Type::OVERALL);
			if (!graph::isomorphism::Isomorphism::isomorphism(ug1, ug2))
				std::cout << "FAIL: isomorphism returned false!" << std::endl;
			measurements::end();
			std::clog << measurements::results() << std::endl;
		}

		return 0;
	} catch(const exception::AlibException& exception) {
		alib::XmlDataFactory::toStdout(exception);
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		std::cerr << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
