/*
 * ExactPatternMatchingAutomaton.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#include "ExactPatternMatchingAutomaton.h"
#include "ExactSubtreeMatchingAutomaton.h"
#include "SubtreeJumpTable.h"

#include <exception/AlibException.h>
#include <tree/ranked/RankedTree.h>
#include <tree/ranked/RankedPattern.h>
#include <tree/ranked/PrefixRankedTree.h>
#include <tree/ranked/PrefixRankedPattern.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedBarPattern.h>

#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/TA/NFTA.h>

#include <alphabet/BottomOfTheStackSymbol.h>

#include <deque>

namespace arbology {

namespace exact {

automaton::Automaton ExactPatternMatchingAutomaton::construct ( const tree::Tree & pattern ) {
	return getInstance ( ).dispatch ( pattern.getData ( ) );
}

automaton::InputDrivenNPDA ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedTree & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonPrefixRankedTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::InputDrivenNPDA, tree::PrefixRankedTree > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

std::vector < alphabet::Symbol > computeRHS ( const tree::PrefixRankedPattern & pattern, const std::vector < int > & patternSubtreeJumpTable, int i ) {
	const std::vector < alphabet::RankedSymbol > & content = pattern.getContent ( );

	unsigned rank = content[i].getRank ( ).getData ( );

	i++;

	std::vector < alphabet::Symbol > res;

	for ( unsigned ranki = 0; ranki < rank; ranki++ ) {
		if ( content[i] == pattern.getSubtreeWildcard ( ) ) {
			res.push_back ( alphabet::symbolFrom ( 'R' ) );
			i++;
		} else {
			res.push_back ( alphabet::symbolFrom ( 'T' ) );

			i = patternSubtreeJumpTable[i];
		}
	}

	return res;
}

automaton::NPDA ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedPattern & pattern ) {
	automaton::NPDA res ( automaton::State ( 0 ), alphabet::symbolFrom ( 'T' ) );

	for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
		if ( symbol == pattern.getSubtreeWildcard ( ) ) continue;

		res.addInputSymbol ( alphabet::Symbol { symbol } );
	}

	res.setStackAlphabet ( { alphabet::symbolFrom ( 'T' ), alphabet::symbolFrom ( 'R' ) } );

	for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
		if ( symbol == pattern.getSubtreeWildcard ( ) ) continue;

		res.addTransition ( automaton::State ( 0 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'T' ) }, automaton::State ( 0 ), std::vector < alphabet::Symbol > { symbol.getRank ( ).getData ( ), alphabet::symbolFrom ( 'T' ) } );
	}

	std::vector < int > patternSubtreeJumpTable = SubtreeJumpTable::compute ( pattern );

	int i = 1;

	for ( const alphabet::RankedSymbol & symbol : pattern.getContent ( ) ) {
		res.addState ( automaton::State ( i ) );

		if ( symbol == pattern.getSubtreeWildcard ( ) )
			for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
				if ( symbol == pattern.getSubtreeWildcard ( ) ) continue;

				if ( symbol.getRank ( ).getData ( ) == 0 ) {
					res.addTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'T' ) }, automaton::State ( i - 1 ), std::vector < alphabet::Symbol > { } );

					res.addTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'R' ) }, automaton::State ( i ), std::vector < alphabet::Symbol > { } );
				} else {
					std::vector < alphabet::Symbol > push {
						symbol.getRank ( ).getData ( ), alphabet::symbolFrom ( 'T' )
					};
					res.addTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'T' ) }, automaton::State ( i - 1 ), push );

					push[symbol.getRank ( ).getData ( ) - 1] = alphabet::symbolFrom ( 'R' );
					res.addTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'R' ) }, automaton::State ( i - 1 ), push );
				}
			}

		else
			res.addTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, std::vector < alphabet::Symbol > { 1, alphabet::symbolFrom ( 'T' ) }, automaton::State ( i ), computeRHS ( pattern, patternSubtreeJumpTable, i - 1 ) );

		i++;
	}

	res.addFinalState ( automaton::State ( i - 1 ) );
	return res;
}

auto ExactPatternMatchingAutomatonPrefixRankedPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NPDA, tree::PrefixRankedPattern > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

automaton::InputDrivenNPDA ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedBarTree & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonPrefixRankedBarTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::InputDrivenNPDA, tree::PrefixRankedBarTree > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

automaton::VisiblyPushdownNPDA ExactPatternMatchingAutomaton::construct ( const tree::PrefixRankedBarPattern & pattern ) {
	automaton::VisiblyPushdownNPDA res ( alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK } );

	res.addState ( automaton::State ( 0 ) );
	res.addInitialState ( automaton::State ( 0 ) );

	for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
		if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) ) continue;

		if ( symbol.getSymbol ( ) == pattern.getBarSymbol ( ) )
			res.addReturnInputSymbol ( alphabet::Symbol { symbol } );
		else
			res.addCallInputSymbol ( alphabet::Symbol { symbol } );
	}

	res.setStackAlphabet ( { alphabet::Symbol { alphabet::BottomOfTheStackSymbol::BOTTOM_OF_THE_STACK }, alphabet::symbolFrom ( 'T' ), alphabet::symbolFrom ( 'R' ) } );

	for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
		if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) ) continue;

		if ( symbol.getSymbol ( ) == pattern.getBarSymbol ( ) )
			res.addReturnTransition ( automaton::State ( 0 ), alphabet::Symbol { symbol }, alphabet::symbolFrom ( 'T' ), automaton::State ( 0 ) );
		else
			res.addCallTransition ( automaton::State ( 0 ), alphabet::Symbol { symbol }, automaton::State ( 0 ), alphabet::symbolFrom ( 'T' ) );
	}

	int i = 1;

	for ( const alphabet::RankedSymbol & symbol : pattern.getContent ( ) ) {
		res.addState ( automaton::State ( i ) );

		if ( symbol == pattern.getSubtreeWildcard ( ) ) {
			for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
				if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) || ( symbol.getSymbol ( ) == pattern.getBarSymbol ( ) ) ) continue;

				res.addCallTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, automaton::State ( i ), alphabet::symbolFrom ( 'R' ) );
			}
		} else if ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) {
			for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
				if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) ) continue;

				if ( symbol.getSymbol ( ) == pattern.getBarSymbol ( ) )
					res.addReturnTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, alphabet::symbolFrom ( 'T' ), automaton::State ( i - 1 ) );
				else
					res.addCallTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, automaton::State ( i - 1 ), alphabet::symbolFrom ( 'T' ) );
			}

			for ( const alphabet::RankedSymbol & symbol : pattern.getAlphabet ( ) ) {
				if ( ( symbol == pattern.getSubtreeWildcard ( ) ) || ( symbol.getSymbol ( ) == pattern.getVariablesBarSymbol ( ) ) || ( symbol.getSymbol ( ) != pattern.getBarSymbol ( ) ) ) continue;

				res.addReturnTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, alphabet::symbolFrom ( 'R' ), automaton::State ( i ) );
			}
		} else if ( symbol.getSymbol ( ) == pattern.getBarSymbol ( ) ) {
			res.addReturnTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, alphabet::symbolFrom ( 'T' ), automaton::State ( i ) );
		} else {
			res.addCallTransition ( automaton::State ( i - 1 ), alphabet::Symbol { symbol }, automaton::State ( i ), alphabet::symbolFrom ( 'T' ) );
		}

		i++;
	}

	res.addFinalState ( automaton::State ( i - 1 ) );
	return res;
}

auto ExactPatternMatchingAutomatonPrefixRankedBarPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::VisiblyPushdownNPDA, tree::PrefixRankedBarPattern > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

automaton::NFTA ExactPatternMatchingAutomaton::construct ( const tree::RankedTree & pattern ) {
	return ExactSubtreeMatchingAutomaton::construct ( pattern );
}

auto ExactPatternMatchingAutomatonRankedTree = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NFTA, tree::RankedTree > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

automaton::State constructRecursivePattern ( const tree::RankedNode & node, automaton::NFTA & res, const alphabet::RankedSymbol & subtreeWildcard, const automaton::State & loopState, int & nextState ) {
	if ( node.getSymbol ( ) == subtreeWildcard ) {
		automaton::State state ( nextState++ );
		res.addState ( state );

		for ( const alphabet::RankedSymbol & symbol : res.getInputAlphabet ( ) ) {
			std::vector < automaton::State > states;
			states.reserve ( symbol.getRank ( ).getData ( ) );

			for ( unsigned i = 0; i < symbol.getRank ( ).getData ( ); i++ )
				states.push_back ( loopState );

			res.addTransition ( symbol, states, state );
		}

		return state;
	} else {
		std::vector < automaton::State > states;
		states.reserve ( node.getSymbol ( ).getRank ( ).getData ( ) );

		for ( const auto & child : node.getChildren ( ) )
			states.push_back ( constructRecursivePattern ( * child, res, subtreeWildcard, loopState, nextState ) );

		automaton::State state ( nextState++ );
		res.addState ( state );
		res.addTransition ( node.getSymbol ( ), states, state );
		return state;
	}
}

automaton::NFTA ExactPatternMatchingAutomaton::construct ( const tree::RankedPattern & pattern ) {
	std::set < alphabet::RankedSymbol > alphabet = pattern.getAlphabet ( );

	alphabet.erase ( pattern.getSubtreeWildcard ( ) );

	automaton::NFTA res;
	res.setInputSymbols ( alphabet );

	int nextState = 0;

	automaton::State loopState ( nextState++ );
	res.addState ( loopState );

	for ( const alphabet::RankedSymbol & symbol : res.getInputAlphabet ( ) ) {
		std::vector < automaton::State > states;
		states.reserve ( symbol.getRank ( ).getData ( ) );

		for ( unsigned i = 0; i < symbol.getRank ( ).getData ( ); i++ )
			states.push_back ( loopState );

		res.addTransition ( symbol, states, loopState );
	}

	res.addFinalState ( constructRecursivePattern ( pattern.getRoot ( ), res, pattern.getSubtreeWildcard ( ), loopState, nextState ) );
	return res;
}

auto ExactPatternMatchingAutomatonRankedPattern = ExactPatternMatchingAutomaton::RegistratorWrapper < automaton::NFTA, tree::RankedPattern > ( ExactPatternMatchingAutomaton::getInstance ( ), ExactPatternMatchingAutomaton::construct );

} /* namespace exact */

} /* namespace arbology */
