/*
 * AllEpsilonClosure.cpp
 *
 *  Created on: 12. 4. 2015
 *	  Author: Jan Travnicek
 */

#include "AllEpsilonClosure.h"

#include <exception/AlibException.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <deque>
#include <iostream>

#include "../../regexp/properties/RegExpEpsilon.h"


namespace automaton {

namespace properties {

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::EpsilonNFA & fsm) {
	std::deque<std::map<automaton::State, std::set<automaton::State>>> Qi;

	Qi.push_back({});
	for(const automaton::State& state : fsm.getStates())
		Qi.at(0)[state].insert(state);

	int i = 1;
	while(true) {
		Qi.push_back( Qi.at( i - 1 ) );

		for( const std::pair<automaton::State, std::set<automaton::State>> & p : Qi.at( i - 1 ) )
			for(const automaton::State& pclosure : p.second )
				for(const auto& transition : fsm.getEpsilonTransitionsFromState(pclosure)) {
					Qi.at( i )[p.first].insert(transition.second.begin(), transition.second.end());
				}

		if( Qi.at( i ) == Qi.at( i - 1 ) )
			break;

		i = i + 1;
	};

	return Qi.at(i);
}

auto AllEpsilonClosureEpsilonNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::EpsilonNFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::MultiInitialStateNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureMultiInitialStateNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::MultiInitialStateNFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::NFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::NFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::DFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureDFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::DFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::ExtendedNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;
	std::map<automaton::State, std::set<automaton::State>> step;

	for(const std::pair<std::pair<State, regexp::RegExp>, std::set<State> >& transition : fsm.getTransitions() )
		if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
			step[transition.first.first].insert(transition.second.begin(), transition.second.end());

	for(const automaton::State& state : fsm.getStates())
		step[state].insert(state);

	do {
		res = step;

		for(const std::pair<automaton::State, std::set<automaton::State>>& item : res)
			for(const automaton::State& to : item.second)
				step[item.first].insert(res[to].begin(), res[to].end());
	} while(res != step);

	return res;
}

auto AllEpsilonClosureExtendedNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::ExtendedNFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::CompactNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;
	std::map<automaton::State, std::set<automaton::State>> step;

	for(const std::pair<std::pair<State, string::LinearString>, std::set<State> >& transition : fsm.getTransitions() )
		if( transition.first.second.getContent().size() == 0 )
			step[transition.first.first].insert(transition.second.begin(), transition.second.end());

	for(const automaton::State& state : fsm.getStates())
		step[state].insert(state);

	do {
		res = step;

		for(const std::pair<automaton::State, std::set<automaton::State>>& item : res)
			for(const automaton::State& to : item.second)
				step[item.first].insert(res[to].begin(), res[to].end());
	} while(res != step);

	return res;
}

auto AllEpsilonClosureCompactNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::CompactNFA>(AllEpsilonClosure::getInstance(), AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure(const Automaton& automaton) {
	return getInstance().dispatch(automaton.getData());
}

} /* namespace properties */

} /* namespace automaton */

