/*
 * Accept.cpp
 *
 *  Created on: 9. 2. 2014
 *	  Author: Jan Travnicek
 */

#include "Run.h"
#include <exception/AlibException.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/DPDA.h>
#include <common/GlobalData.h>

#include <deque>

namespace automaton {

namespace run {

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, State, std::set < unsigned > > Run::calculateState ( const automaton::DFA & automaton, const string::LinearString & string ) {
	bool res = true;
	unsigned i = 0;
	std::set < unsigned > occurrences;
	automaton::State state = automaton.getInitialState ( );

	if ( automaton.getFinalStates ( ).count ( state ) )
		occurrences.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const alphabet::Symbol & symbol : string.getContent ( ) ) {
		auto transition = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

		if ( transition == automaton.getTransitions ( ).end ( ) ) {
			res = false;
			break;
		}

		state = transition->second;
		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occurrences.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occurrences );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, std::set < State >, std::set < unsigned > > Run::calculateStates ( const automaton::NFA & automaton, const string::LinearString & string ) {
	bool res = true;
	unsigned i = 0;
	std::set < unsigned > occurrences;
	std::set < automaton::State > states {
		automaton.getInitialState ( )
	};

	for ( const State & state : states )
		if ( automaton.getFinalStates ( ).count ( state ) )
			occurrences.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << states << std::endl;

	for ( const alphabet::Symbol & symbol : string.getContent ( ) ) {
		std::set < State > next;

		for ( const State & state : states ) {
			auto transitions = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transitions == automaton.getTransitions ( ).end ( ) ) continue;

			next.insert ( transitions->second.begin ( ), transitions->second.end ( ) );
		}

		i++;
		states = next;

		for ( const State & state : states )
			if ( automaton.getFinalStates ( ).count ( state ) )
				occurrences.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << states << std::endl;
	}

	return std::make_tuple ( res, states, occurrences );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::pair < bool, State > Run::calculateState ( const automaton::DFTA & automaton, const tree::RankedNode & node, std::set < unsigned > & occ, unsigned & i ) {
	std::vector < State > states;

	states.reserve ( node.getSymbol ( ).getRank ( ).getData ( ) );

	unsigned tmp = i;
	i++;

	bool sign = true;

	for ( const auto & child : node.getChildren ( ) ) {
		std::pair < bool, State > res = calculateState ( automaton, * child, occ, i );

		if ( res.first == false )
			sign = false;
		else
			states.push_back ( res.second );
	}

	if ( !sign ) return std::make_pair ( false, automaton::State ( "fail" ) );

	const auto & it = automaton.getTransitions ( ).find ( std::make_pair ( node.getSymbol ( ), states ) );

	if ( it == automaton.getTransitions ( ).end ( ) ) return std::make_pair ( false, automaton::State ( "fail" ) );

	State state = it->second;

	if ( automaton.getFinalStates ( ).count ( state ) ) occ.insert ( tmp );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	return std::make_pair ( true, state );
}

std::tuple < bool, State, std::set < unsigned > > Run::calculateState ( const automaton::DFTA & automaton, const tree::RankedTree & tree ) {
	std::set < unsigned > occ;
	unsigned i = 0;
	std::pair < bool, State > res = calculateState ( automaton, tree.getRoot ( ), occ, i );

	return std::make_tuple ( res.first, res.second, occ );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::pair < bool, std::set < State > > Run::calculateStates ( const automaton::NFTA & automaton, const tree::RankedNode & node, std::set < unsigned > & occ, unsigned & i ) {
	std::vector < std::set < State > > resStates;

	resStates.reserve ( node.getSymbol ( ).getRank ( ).getData ( ) );

	unsigned tmp = i;
	i++;

	bool sign = true;

	for ( const auto & child : node.getChildren ( ) ) {
		std::pair < bool, std::set < State > > childStates = calculateStates ( automaton, * child, occ, i );

		if ( childStates.second.empty ( ) )
			sign = false;

		resStates.push_back ( childStates.second );
	}

	std::set < State > states;

	for ( const auto & transition : automaton.getTransitions ( ) ) {
		if ( transition.first.first != node.getSymbol ( ) ) continue;

		unsigned rank = transition.first.first.getRank ( ).getData ( );

		for ( unsigned i = 0; i < rank; i++ )
			if ( !resStates[i].count ( transition.first.second[i] ) ) break;

		if ( i == rank ) states.insert ( transition.second.begin ( ), transition.second.end ( ) );
	}

	for ( const State & state : states )
		if ( automaton.getFinalStates ( ).count ( state ) ) occ.insert ( tmp );

	if ( common::GlobalData::verbose )
		std::clog << states << std::endl;

	return std::make_pair ( sign, states );
}

std::tuple < bool, std::set < State >, std::set < unsigned > > Run::calculateStates ( const automaton::NFTA & automaton, const tree::RankedTree & tree ) {
	std::set < unsigned > occ;
	unsigned i = 0;
	std::pair < bool, std::set < State > > res = calculateStates ( automaton, tree.getRoot ( ), occ, i );

	return std::make_tuple ( res.first, res.second, occ );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

bool Run::canPop ( const std::deque < alphabet::Symbol > & pushdownStore, const std::vector < alphabet::Symbol > & pop ) {
	unsigned popSize = pop.size ( );
	unsigned pushdownStoreSize = pushdownStore.size ( );

	if ( pushdownStoreSize < popSize ) return false;

	for ( unsigned i = 0; i < popSize; i++ )
		if ( pop[i] != pushdownStore[pushdownStoreSize - i - 1] )
			return false;

	return true;
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, State, std::set < unsigned >, std::deque < alphabet::Symbol > > Run::calculateState ( const automaton::InputDrivenDPDA & automaton, const string::LinearString & string ) {
	automaton::State state = automaton.getInitialState ( );
	std::deque < alphabet::Symbol > pushdownStore {
		automaton.getInitialSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const alphabet::Symbol & symbol : string.getContent ( ) ) {
		auto transition = automaton.getTransitions ( ).find ( std::make_pair ( state, symbol ) );

		if ( transition == automaton.getTransitions ( ).end ( ) ) {
			res = false;
			break;
		}

		const std::pair < std::vector < alphabet::Symbol >, std::vector < alphabet::Symbol > > & operation = automaton.getPushdownStoreOperations ( ).find ( symbol )->second;

		if ( !canPop ( pushdownStore, operation.first ) ) {
			res = false;
			break;
		}

		for ( unsigned j = 0; j < operation.first.size ( ); j++ ) pushdownStore.pop_back ( );

		for ( auto iter = operation.second.rbegin ( ); iter != operation.second.rend ( ); ++iter ) pushdownStore.push_back ( * iter );

		state = transition->second;
		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, State, std::set < unsigned >, std::deque < alphabet::Symbol > > Run::calculateState ( const automaton::VisiblyPushdownDPDA & automaton, const string::LinearString & string ) {
	automaton::State state = automaton.getInitialState ( );
	std::deque < alphabet::Symbol > pushdownStore {
		automaton.getBottomOfTheStackSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( const alphabet::Symbol & symbol : string.getContent ( ) ) {
		if ( automaton.getCallInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transition == automaton.getCallTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			pushdownStore.push_back ( transition->second.second );
			state = transition->second.first;
		} else if ( automaton.getReturnInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, symbol, pushdownStore.back ( ) ) );

			if ( transition == automaton.getReturnTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			if ( pushdownStore.back ( ) != automaton.getBottomOfTheStackSymbol ( ) ) pushdownStore.pop_back ( );

			state = transition->second;
		} else if ( automaton.getLocalInputAlphabet ( ).count ( symbol ) ) {
			auto transition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, symbol ) );

			if ( transition == automaton.getLocalTransitions ( ).end ( ) ) {
				res = false;
				break;
			}

			state = transition->second;
		} else {
			res = false;
			break;
		}

		i++;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, State, std::set < unsigned >, std::deque < alphabet::Symbol > > Run::calculateState ( const automaton::RealTimeHeightDeterministicDPDA & automaton, const string::LinearString & string ) {
	automaton::State state = automaton.getInitialState ( );
	std::deque < alphabet::Symbol > pushdownStore {
		automaton.getBottomOfTheStackSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose )
		std::clog << state << std::endl;

	for ( auto symbolIter = string.getContent ( ).begin ( ); symbolIter != string.getContent ( ).end ( ); ) {

		auto callTransition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, * symbolIter ) );

		if ( callTransition != automaton.getCallTransitions ( ).end ( ) ) {
			symbolIter++;
			i++;
		} else {
			callTransition = automaton.getCallTransitions ( ).find ( std::make_pair ( state, string::Epsilon::EPSILON ) );
		}

		if ( callTransition != automaton.getCallTransitions ( ).end ( ) ) {
			pushdownStore.push_back ( callTransition->second.second );
			state = callTransition->second.first;
		} else {
			res = false;
		}

		auto returnTransition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, * symbolIter, pushdownStore.back ( ) ) );

		if ( returnTransition != automaton.getReturnTransitions ( ).end ( ) ) {
			symbolIter++;
			i++;
		} else {
			returnTransition = automaton.getReturnTransitions ( ).find ( std::make_tuple ( state, string::Epsilon::EPSILON, pushdownStore.back ( ) ) );
		}

		if ( returnTransition != automaton.getReturnTransitions ( ).end ( ) ) {
			pushdownStore.pop_back ( );
			state = returnTransition->second;
		} else {
			res = false;
		}

		auto localTransition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, * symbolIter ) );

		if ( localTransition != automaton.getLocalTransitions ( ).end ( ) ) {
			symbolIter++;
			i++;
		} else {
			localTransition = automaton.getLocalTransitions ( ).find ( std::make_pair ( state, string::Epsilon::EPSILON ) );
		}

		if ( localTransition != automaton.getLocalTransitions ( ).end ( ) )
			state = localTransition->second;
		else
			res = false;

		if ( res == false )
			break;

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose )
			std::clog << state << std::endl;
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

std::tuple < bool, State, std::set < unsigned >, std::deque < alphabet::Symbol > > Run::calculateState ( const automaton::DPDA & automaton, const string::LinearString & string ) {
	automaton::State state = automaton.getInitialState ( );
	std::deque < alphabet::Symbol > pushdownStore {
		automaton.getInitialSymbol ( )
	};
	unsigned i = 0;
	std::set < unsigned > occ;
	bool res = true;

	if ( automaton.getFinalStates ( ).count ( state ) )
		occ.insert ( i );

	if ( common::GlobalData::verbose ) {
		std::clog << "State : " << state << std::endl;
		std::clog << "PushdownStore : " << pushdownStore << std::endl;
	}

	for ( auto symbolIter = string.getContent ( ).begin ( ); symbolIter != string.getContent ( ).end ( ); ) {
		auto transitions = automaton.getTransitionsFromState ( state );
		auto transition	 = transitions.begin ( );

		for ( ; transition != transitions.end ( ); transition++ ) {
			if ( ( std::get < 1 > ( transition->first ) != * symbolIter ) && !std::get < 1 > ( transition->first ).is < string::Epsilon > ( ) ) continue;

			if ( canPop ( pushdownStore, std::get < 2 > ( transition->first ) ) ) break;
		}

		if ( transition == transitions.end ( ) ) {
			res = false;
			break;
		}

		if ( common::GlobalData::verbose )
			std::clog << "Transition: " << * transition << std::endl;

		for ( unsigned j = 0; j < std::get < 2 > ( transition->first ).size ( ); j++ ) pushdownStore.pop_back ( );

		for ( auto iter = transition->second.second.rbegin ( ); iter != transition->second.second.rend ( ); ++iter ) pushdownStore.push_back ( * iter );

		state = transition->second.first;

		if ( !std::get < 1 > ( transition->first ).is < string::Epsilon > ( ) ) {
			i++;
			symbolIter++;
		}

		if ( automaton.getFinalStates ( ).count ( state ) )
			occ.insert ( i );

		if ( common::GlobalData::verbose ) {
			std::clog << "State: " << state << std::endl;
			std::clog << "PushdownStore: " << pushdownStore << std::endl;
		}
	}

	return std::make_tuple ( res, state, occ, pushdownStore );
}

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

} /* namespace run */

} /* namespace automaton */
