/*
 * LeftRecursionRemover.cpp
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#include "LeftRecursionRemover.h"

#include "../convert/ToGrammarRightRG.h"
#include <vector>

namespace grammar {

namespace simplify {

grammar::EpsilonFreeCFG directLeftRecursionRemoveAsOrder(const grammar::EpsilonFreeCFG& origGrammar) {
	grammar::EpsilonFreeCFG res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const auto& nonterminal : origGrammar.getNonterminalAlphabet()) {
		if(origGrammar.getRules().find(nonterminal) == origGrammar.getRules().end()) continue;

		if(std::any_of(origGrammar.getRules().find(nonterminal)->second.begin(), origGrammar.getRules().find(nonterminal)->second.end(), [&](const std::vector<alphabet::Symbol>& singleRHS) {
				return singleRHS[0] == nonterminal; // is there a direct left recursion?
			} ) && std::all_of(origGrammar.getRules().find(nonterminal)->second.begin(), origGrammar.getRules().find(nonterminal)->second.end(), [&](const std::vector<alphabet::Symbol>& singleRHS) {
				return origGrammar.getTerminalAlphabet().count(singleRHS[0]) || singleRHS[0] >= nonterminal; // only remove left recursion when all nonterminals are bigger than the left hand side
			})) {
			alphabet::Symbol primed = alphabet::createUniqueSymbol(nonterminal, res.getTerminalAlphabet(), res.getNonterminalAlphabet());
			res.addNonterminalSymbol(primed);
			for(const std::vector<alphabet::Symbol>& singleRHS : origGrammar.getRules().find(nonterminal)->second) { // do the removal
				if(singleRHS[0] == nonterminal) { // A -> A alpha
					std::vector<alphabet::Symbol> tmpRHS(singleRHS.begin() + 1, singleRHS.end());

					res.addRule(primed, tmpRHS); // A' -> alpha

					tmpRHS.push_back(primed);
					res.addRule(primed, tmpRHS); // A' -> alpha A'
				} else { // a -> beta
					std::vector<alphabet::Symbol> tmpRHS(singleRHS);

					res.addRule(nonterminal, tmpRHS); // A -> beta

					tmpRHS.push_back(primed);
					res.addRule(nonterminal, tmpRHS); // A -> beta A'
				}
			}
		} else {
			for(const std::vector<alphabet::Symbol>& singleRHS : origGrammar.getRules().find(nonterminal)->second) {
				res.addRule(nonterminal, singleRHS);
			}
		}
	}
	return res;
}

grammar::EpsilonFreeCFG assignAsOrder(const grammar::EpsilonFreeCFG& origGrammar, unsigned i, const std::set<alphabet::Symbol>& origNonterminals) {
	grammar::EpsilonFreeCFG res(origGrammar.getInitialSymbol());
	res.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	res.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	res.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());

	for(const alphabet::Symbol& lhs : origGrammar.getNonterminalAlphabet()) {
		if(i > 0) {
			if(origGrammar.getRules().find(lhs) == origGrammar.getRules().end()) continue;
			for(const std::vector<alphabet::Symbol>& rule : origGrammar.getRules().find(lhs)->second) {
				res.addRule(lhs, rule);
			}

			i--;
			continue; // substitue only in i-th up to n-th nonterminals
		}
		if(origGrammar.getRules().find(lhs) == origGrammar.getRules().end()) continue;
		if(!origNonterminals.count(lhs)) { // do not subsitute in nonoriginal nonterminals
			for(const std::vector<alphabet::Symbol>& rule : origGrammar.getRules().find(lhs)->second) {
				res.addRule(lhs, rule);
			}
			continue;
		}

		const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule = *origGrammar.getRules().find(lhs);

		for(const std::vector<alphabet::Symbol>& singleRHS : rule.second) {
			if(res.getTerminalAlphabet().count(singleRHS[0])) { //do not substitute terminals
				res.addRule(lhs, singleRHS);
				continue;
			}
			const alphabet::Symbol& secondLHS = singleRHS[0];
			if(secondLHS >= lhs) { // substitute only by 0th up to i-th nonterminals right hand sides
				res.addRule(lhs, singleRHS);
				continue;
			}
			if(origGrammar.getRules().find(secondLHS) == origGrammar.getRules().end()) { //is there any right hand side to substitue with?
				//if not well this rule does not generate anything anyway
				continue;
			}

			for(const std::vector<alphabet::Symbol>& secondSingleRHS : origGrammar.getRules().find(secondLHS)->second) { // do the substitution
				std::vector<alphabet::Symbol> newRHS(secondSingleRHS);
				newRHS.insert(newRHS.end(), singleRHS.begin() + 1, singleRHS.end());
				res.addRule(lhs, newRHS);
			}
		}
	}
	return res;
}

grammar::EpsilonFreeCFG LeftRecursionRemover::remove(const grammar::EpsilonFreeCFG& origGrammar) {
	grammar::EpsilonFreeCFG step = origGrammar;
	for(const alphabet::Symbol& nonterminal : step.getNonterminalAlphabet()) { // remove identities
		step.removeRule(nonterminal, std::vector<alphabet::Symbol>{nonterminal});
	}
	unsigned i = 0;
	while(i < origGrammar.getNonterminalAlphabet().size()) {
		grammar::EpsilonFreeCFG nextStep = assignAsOrder(directLeftRecursionRemoveAsOrder(step), i, origGrammar.getNonterminalAlphabet());

		if(step == nextStep) break;
		step = std::move(nextStep);
		i++;
	};

	return step;
}

auto LeftRecursionRemoverEpsilonFreeCFG = LeftRecursionRemover::RegistratorWrapper<grammar::EpsilonFreeCFG, grammar::EpsilonFreeCFG>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::EpsilonFreeCFG LeftRecursionRemover::remove(const grammar::CNF& origGrammar) {
	EpsilonFreeCFG tmp(origGrammar.getInitialSymbol());
	tmp.setTerminalAlphabet(origGrammar.getTerminalAlphabet());
	tmp.setNonterminalAlphabet(origGrammar.getNonterminalAlphabet());
	tmp.setGeneratesEpsilon(origGrammar.getGeneratesEpsilon());
	for(const auto& rule : origGrammar.getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.is<alphabet::Symbol>()) {
				tmp.addRule(rule.first, {rhs.get<alphabet::Symbol>()});
			} else {
				const auto& rhsPair = rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				tmp.addRule(rule.first, {rhsPair.first, rhsPair.second});
			}
		}
	}
	return remove(tmp);
}

auto LeftRecursionRemoverCNF = LeftRecursionRemover::RegistratorWrapper<grammar::EpsilonFreeCFG, grammar::CNF>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::GNF LeftRecursionRemover::remove(const grammar::GNF& origGrammar) {
	return origGrammar;
}

auto LeftRecursionRemoverGNF = LeftRecursionRemover::RegistratorWrapper<grammar::GNF, grammar::GNF>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::RightRG LeftRecursionRemover::remove(const grammar::RightRG& origGrammar) {
	return origGrammar;
}

auto LeftRecursionRemoverRightRG = LeftRecursionRemover::RegistratorWrapper<grammar::RightRG, grammar::RightRG>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::RightLG LeftRecursionRemover::remove(const grammar::RightLG& origGrammar) {
	return origGrammar;
}

auto LeftRecursionRemoverRightLG = LeftRecursionRemover::RegistratorWrapper<grammar::RightLG, grammar::RightLG>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::RightRG LeftRecursionRemover::remove(const grammar::LeftRG& origGrammar) {
	return convert::ToGrammarRightRG::convert(origGrammar);
}

auto LeftRecursionRemoverLeftRG = LeftRecursionRemover::RegistratorWrapper<grammar::RightRG, grammar::LeftRG>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::RightLG LeftRecursionRemover::remove(const grammar::LeftLG& /* origGrammar */) {
	throw exception::AlibException("LeftRecursionRemover: Removing from LeftLG NYI"); // TODO
}

auto LeftRecursionRemoverLeftLG = LeftRecursionRemover::RegistratorWrapper<grammar::RightLG, grammar::LeftLG>(LeftRecursionRemover::getInstance(), LeftRecursionRemover::remove);

grammar::Grammar LeftRecursionRemover::remove(const grammar::Grammar& grammar) {
	return getInstance().dispatch(grammar.getData());
}

} /* namespace simplify */

} /* namespace grammar */
