/*
 * ToAutomatonThompson.h
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#ifndef TO_AUTOMATON_THOMPSON_H_
#define TO_AUTOMATON_THOMPSON_H_

#include <common/multipleDispatch.hpp>

#include <regexp/RegExp.h>
#include <regexp/formal/FormalRegExp.h>
#include <regexp/unbounded/UnboundedRegExp.h>
#include <automaton/FSM/EpsilonNFA.h>

namespace regexp {

namespace convert {

/**
 * Converts regular expression to finite automaton using Thompson's Construction Algorithm (TCA).
 * Sources:
 *  Hopcroft, section 3.2.3
 *  http://www.eecis.udel.edu/~cavazos/cisc672/lectures/Lecture-04.pdf
 *  http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.21.7450&rep=rep1&type=ps
 *  Melichar 2.112
 */
class ToAutomatonThompson : public std::SingleDispatch<automaton::EpsilonNFA, regexp::RegExpBase>, regexp::FormalRegExpElement::const_visitor_type, regexp::UnboundedRegExpElement::const_visitor_type {
public:
	ToAutomatonThompson() {}

	/**
	 * Performs conversion.
	 * @param regexp regexp to convert
	 * @return nondeterministic finite automaton with epsilon transitions accepting language described by the regexp
	 */
	static automaton::EpsilonNFA convert(const regexp::RegExp& regexp);

	template<class T>
	static automaton::EpsilonNFA convert(const T& regexp);

private:
	 void Visit(void*, const regexp::UnboundedRegExpAlternation& alternation) const;
	 void Visit(void*, const regexp::UnboundedRegExpConcatenation& concatenation) const;
	 void Visit(void*, const regexp::UnboundedRegExpIteration& iteration) const;
	 void Visit(void*, const regexp::UnboundedRegExpSymbol& symbol) const;
	 void Visit(void*, const regexp::UnboundedRegExpEpsilon& epsilon) const;
	 void Visit(void*, const regexp::UnboundedRegExpEmpty& empty) const;

	 void Visit(void*, const regexp::FormalRegExpAlternation& alternation) const;
	 void Visit(void*, const regexp::FormalRegExpConcatenation& concatenation) const;
	 void Visit(void*, const regexp::FormalRegExpIteration& iteration) const;
	 void Visit(void*, const regexp::FormalRegExpSymbol& symbol) const;
	 void Visit(void*, const regexp::FormalRegExpEpsilon& epsilon) const;
	 void Visit(void*, const regexp::FormalRegExpEmpty& empty) const;

	 static const ToAutomatonThompson TO_AUTOMATON_THOMPSON;

public:
	static ToAutomatonThompson& getInstance() {
		static ToAutomatonThompson res;
		return res;
	}
};

} /* namespace convert */

} /* namespace regexp */

#endif /* TO_AUTOMATON_THOMPSON_H_ */
