/*
 * InputDrivenNPDA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "InputDrivenNPDA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>
#include <algorithm>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace automaton {

InputDrivenNPDA::InputDrivenNPDA(State initialState, alphabet::Symbol initialPushdownSymbol) : SingleInitialSymbolPushdownStoreAlphabet(std::move(initialPushdownSymbol)), SingleInitialState(std::move(initialState)) {

}

AutomatonBase* InputDrivenNPDA::clone() const {
	return new InputDrivenNPDA(*this);
}

AutomatonBase* InputDrivenNPDA::plunder() && {
	return new InputDrivenNPDA(std::move(*this));
}

bool InputDrivenNPDA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.first.first == state || transition.second.find(state) != transition.second.end())
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool InputDrivenNPDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.first.second == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool InputDrivenNPDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (const auto& pushdownStoreOperation : inputSymbolToPushdownStoreOperation) {
		if (std::find(pushdownStoreOperation.second.first.begin(), pushdownStoreOperation.second.first.end(), symbol) != pushdownStoreOperation.second.first.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		if (std::find(pushdownStoreOperation.second.second.begin(), pushdownStoreOperation.second.second.end(), symbol) != pushdownStoreOperation.second.second.end())
			throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	if(initialSymbol == symbol) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

void InputDrivenNPDA::checkPushdownStoreOperation(const alphabet::Symbol& input, const std::vector<alphabet::Symbol>& pop, const std::vector<alphabet::Symbol>& push) {
	if (inputAlphabet.find(input) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	for(const alphabet::Symbol& popSymbol : pop) {
		if (stackAlphabet.find(popSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) popSymbol + "\" doesn't exist.");
		}
	}

	for(const alphabet::Symbol& pushSymbol : push) {
		if (stackAlphabet.find(pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) pushSymbol + "\" doesn't exist.");
		}
	}
}

bool InputDrivenNPDA::setPushdownStoreOperation(alphabet::Symbol input, std::vector<alphabet::Symbol> pop, std::vector<alphabet::Symbol> push) {
	checkPushdownStoreOperation(input, pop, push);
	return inputSymbolToPushdownStoreOperation.insert(std::make_pair(input, std::make_pair(pop, push))).second;
}

void InputDrivenNPDA::setPushdownStoreOperations(std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>> operations) {
	std::set<alphabet::Symbol> removed;
	std::set_difference(inputAlphabet.begin(), inputAlphabet.end(), key_begin(operations), key_end(operations), std::inserter(removed, removed.end()));

	for(const alphabet::Symbol& removedSymbol : removed) {
		clearPushdownStoreOperation(removedSymbol);
	}

	for(const auto& added : operations) {
		checkPushdownStoreOperation(added.first, added.second.first, added.second.second);
	}

	inputSymbolToPushdownStoreOperation = std::move(operations);
}

bool InputDrivenNPDA::clearPushdownStoreOperation(const alphabet::Symbol& input) {
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.first.second == input)
			throw AutomatonException("Input symbol \"" + (std::string) input + "\" is used.");
	}

	return inputSymbolToPushdownStoreOperation.erase(input);
}

const std::map<alphabet::Symbol, std::pair<std::vector<alphabet::Symbol>, std::vector<alphabet::Symbol>>>& InputDrivenNPDA::getPushdownStoreOperations() const {
	return inputSymbolToPushdownStoreOperation;
}

bool InputDrivenNPDA::addTransition(State from, alphabet::Symbol input, State to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (inputAlphabet.find(input) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (inputSymbolToPushdownStoreOperation.find(input) == inputSymbolToPushdownStoreOperation.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(std::move(from), std::move(input));

	return transitions[std::move(key)].insert(std::move(to)).second;
}

bool InputDrivenNPDA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);

	return transitions[key].erase(to);
}

const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >& InputDrivenNPDA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > InputDrivenNPDA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsFromState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.first.first == from) {
			transitionsFromState[transition.first].insert(transition.second.begin(), transition.second.end());
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, alphabet::Symbol>, std::set<State> > InputDrivenNPDA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > transitionsToState;
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.second.find(to) != transition.second.end()) {
			transitionsToState[transition.first].insert(transition.second.begin(), transition.second.end());
		}
	}

	return transitionsToState;
}

bool InputDrivenNPDA::isDeterministic() const {
	for (const std::pair<std::pair<State, alphabet::Symbol>, std::set<State> >& transition : transitions) {
		if (transition.second.size() != 1 || transition.second.size() != 0) {
			return false;
		}
	}

	return true;
}

int InputDrivenNPDA::compare(const InputDrivenNPDA& other) const {
	auto first = std::tie(states, inputAlphabet, initialState, finalStates, stackAlphabet, initialSymbol, inputSymbolToPushdownStoreOperation, transitions);
	auto second = std::tie(other.states, other.inputAlphabet, other.initialState, other.finalStates, other.stackAlphabet, other.initialSymbol, other.inputSymbolToPushdownStoreOperation, other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void InputDrivenNPDA::operator>>(std::ostream& out) const {
	out << "(InputDrivenNPDA "
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialState = " << initialState
		<< " finalStates = " << finalStates
		<< " stackAlphabet = " << stackAlphabet
		<< " initialSymbol = " << initialSymbol
		<< " transitions = " << transitions
		<< " inputSymbolToPushdownStoreOperation = " << inputSymbolToPushdownStoreOperation
		<< ")";
}

InputDrivenNPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string InputDrivenNPDA::XML_TAG_NAME = "InputDrivenNPDA";

InputDrivenNPDA InputDrivenNPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, InputDrivenNPDA::XML_TAG_NAME);

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parseStackAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol initialStackSymbol = AutomatonFromXMLParser::parseInitialStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	InputDrivenNPDA automaton(std::move(initialState), std::move(initialStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setStackAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	automaton.setPushdownStoreOperations(AutomatonFromXMLParser::parseInputToPushdownStoreOperation(input));
	AutomatonFromXMLParser::parseTransitions<InputDrivenNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, InputDrivenNPDA::XML_TAG_NAME);
	return automaton;
}

void InputDrivenNPDA::parseTransition(std::deque<sax::Token>::iterator& input, InputDrivenNPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to));
}

void InputDrivenNPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(InputDrivenNPDA::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeStackAlphabet(out, this->getStackAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialStackSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	AutomatonToXMLComposer::composeInputToPushdownStoreOperation(out, this->getPushdownStoreOperations());
	composeTransitions(out);

	out.emplace_back(InputDrivenNPDA::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void InputDrivenNPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto inputDrivenNPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::InputDrivenNPDA>();
auto inputDrivenNPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::InputDrivenNPDA>();

} /* namespace alib */
