/*
 * NPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef NPDA_H_
#define NPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/SingleInitialSymbolPushdownStoreAlphabet.h"
#include "../common/InputAlphabet.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

/**
 * Push Down Automaton
 */
class NPDA : public AutomatonBase, public SingleInitialSymbolPushdownStoreAlphabet, public SingleInitialState, public InputAlphabet {
protected:
	std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::set < std::pair < State, std::vector < alphabet::Symbol > > > > transitions;

public:
	explicit NPDA ( State initialState, alphabet::Symbol initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState ( const State & state );

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol ( const alphabet::Symbol & symbol );

	/**
	 * @copydoc Automaton::removeStackSymbol(const Symbol&)
	 */
	virtual bool removeStackSymbol ( const alphabet::Symbol & symbol );

	/**
	 * Adds transition to the NPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the NPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( State from, std::variant < string::Epsilon, alphabet::Symbol > input, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	bool addTransition ( State from, alphabet::Symbol input, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	bool addTransition ( State from, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	/**
	 * Removes the transition from the NPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the NPDA
	 */
	bool removeTransition ( const State & from, const std::variant < string::Epsilon, alphabet::Symbol > & input, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	bool removeTransition ( const State & from, const alphabet::Symbol & input, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	bool removeTransition ( const State & from, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	/**
	 * @return NPDA transitions
	 */
	const std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::set < std::pair < State, std::vector < alphabet::Symbol > > > > & getTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static NPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, NPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
struct compare < automaton::NPDA > {
	int operator ()( const automaton::NPDA & first, const automaton::NPDA & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* NPDA_H_ */
