/*
 * LeftRG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "LeftRG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

LeftRG::LeftRG(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)), generatesEpsilon(false) {

}

LeftRG::LeftRG(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)), generatesEpsilon(false) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* LeftRG::clone() const {
	return new LeftRG(*this);
}

GrammarBase* LeftRG::plunder() && {
	return new LeftRG(std::move(*this));
}

bool LeftRG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second) {
			if((rhs.is<alphabet::Symbol>() && rhs.get<alphabet::Symbol>() == symbol) || (rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().first == symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
		}
	}

	return terminalAlphabet.erase(symbol);
}

bool LeftRG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second)
			if(rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().second == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");

	return nonterminalAlphabet.erase(symbol);
}

bool LeftRG::addRule(alphabet::Symbol leftHandSide, std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<alphabet::Symbol>()) {
		const alphabet::Symbol& rhs = rightHandSide.get<alphabet::Symbol>();

		if(!terminalAlphabet.count(rhs))
			throw GrammarException("Rule must rewrite to terminal symbol");

		return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
	} else {
		const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = rightHandSide.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();

		if(!nonterminalAlphabet.count(rhs.first) || !terminalAlphabet.count(rhs.second))
			throw GrammarException("Rule must rewrite to terminal symbol followed by nonterminal symbol");

		return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
	}
}

bool LeftRG::addRule(alphabet::Symbol leftHandSide, alphabet::Symbol rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(std::move(rightHandSide));
	return addRule(std::move(leftHandSide), std::move(rhs));
}

bool LeftRG::addRule(alphabet::Symbol leftHandSide, std::pair<alphabet::Symbol, alphabet::Symbol> rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(std::move(rightHandSide));
	return addRule(std::move(leftHandSide), std::move(rhs));
}

const std::map<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>> & LeftRG::getRules() const {
	return rules;
}

bool LeftRG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool LeftRG::removeRule(const alphabet::Symbol& leftHandSide, const alphabet::Symbol& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool LeftRG::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, alphabet::Symbol>& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool LeftRG::addRawRule(alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(true);
		return res;
	} else if(rightHandSide.size() == 1) {
		return addRule(std::move(leftHandSide), std::move(rightHandSide[0]));
	} else if(rightHandSide.size() == 2) {
		return addRule(std::move(leftHandSide), std::make_pair(std::move(rightHandSide[0]), std::move(rightHandSide[1])));
	} else {
		throw exception::AlibException("Invalid right hand side");
	}
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> LeftRG::getRawRules() const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> res;
	for(const auto& rule : getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.is<alphabet::Symbol>()) {
				std::vector<alphabet::Symbol> tmp;
				tmp.push_back(rhs.get<alphabet::Symbol>());
				res[rule.first].insert(tmp);
			} else {
				const auto& rhsPair = rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
				std::vector<alphabet::Symbol> tmp;
				tmp.push_back(rhsPair.first);
				tmp.push_back(rhsPair.second);
				res[rule.first].insert(tmp);
			}
		}
	}
	if(getGeneratesEpsilon()) {
		res[getInitialSymbol()].insert(std::vector<alphabet::Symbol> {});
	}
	return res;
}

bool LeftRG::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		if(leftHandSide != initialSymbol) throw exception::AlibException("Illegal left hand side of epsilon rule");
		bool res = getGeneratesEpsilon();
		setGeneratesEpsilon(false);
		return res;
	} else if(rightHandSide.size() == 1) {
		return removeRule(leftHandSide, rightHandSide[0]);
	} else if(rightHandSide.size() == 2) {
		return removeRule(leftHandSide, std::make_pair(rightHandSide[0], rightHandSide[1]));
	} else {
		throw exception::AlibException("Invalid right hand side");
	}
}

void LeftRG::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool LeftRG::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

int LeftRG::compare(const LeftRG& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void LeftRG::operator>>(std::ostream& out) const {
	out << "(LeftRG "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< "generatesEpsilon = " << generatesEpsilon
			<< ")";
}

LeftRG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string LeftRG::XML_TAG_NAME = "LeftRG";

LeftRG LeftRG::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, LeftRG::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	LeftRG grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon(input);
	grammar.setGeneratesEpsilon(generatesEpsilon);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, LeftRG::XML_TAG_NAME);
	return grammar;
}

void LeftRG::parseRule(std::deque<sax::Token>::iterator& input, LeftRG& grammar) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS(input);
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs = GrammarFromXMLParser::parseRuleOneOrTwoSymbolsRHS(input);
	grammar.addRule(std::move(lhs), std::move(rhs));
}

void LeftRG::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(LeftRG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);
	GrammarToXMLComposer::composeGeneratesEpsilon(out, this->getGeneratesEpsilon());

	out.emplace_back(LeftRG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void LeftRG::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleSingleSymbolLHS(out, rule.first);
			GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

auto LeftRGParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::LeftRG>();
auto LeftRGParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::LeftRG>();

} /* namespace alib */
