/*
 * RightLG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "RightLG.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

RightLG::RightLG(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {

}

RightLG::RightLG(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* RightLG::clone() const {
	return new RightLG(*this);
}

GrammarBase* RightLG::plunder() && {
	return new RightLG(std::move(*this));
}

bool RightLG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> >& rule : rules) {
		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::vector<alphabet::Symbol>>()) {
				const std::vector<alphabet::Symbol>& rhs = rhsTmp.get<std::vector<alphabet::Symbol>>();

				if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			} else {
				const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rhsTmp.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

				if(std::find(rhs.first.begin(), rhs.first.end(), symbol) != rhs.first.end())
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			}
	}

	return terminalAlphabet.erase(symbol);
}

bool RightLG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> >& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>()) {
				const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rhsTmp.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

				if(rhs.second == symbol)
					throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
			}
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool RightLG::addRule(alphabet::Symbol leftHandSide, std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = rightHandSide.get<std::vector<alphabet::Symbol>>();
		for(const auto & symbol : rhs ) {
			if(!terminalAlphabet.count(symbol))
				throw GrammarException("Symbol " + (std::string) symbol + " is not a terminal symbol");
		}

		return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
	} else {
		const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rhs = rightHandSide.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();

		if(!nonterminalAlphabet.count(rhs.second))
			throw GrammarException("Symbol " + (std::string) rhs.second + " is not a nonterminal symbol");

		for(const auto & symbol : rhs.first ) {
			if(!terminalAlphabet.count(symbol))
				throw GrammarException("Symbol " + (std::string) symbol + " is not a terminal symbol");
		}

		return rules[std::move(leftHandSide)].insert(std::move(rightHandSide)).second;
	}
}

bool RightLG::addRule(alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(std::move(rightHandSide));
	return addRule(std::move(leftHandSide), std::move(rhs));
}

bool RightLG::addRule(alphabet::Symbol leftHandSide, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(std::move(rightHandSide));
	return addRule(std::move(leftHandSide), std::move(rhs));
}

const std::map<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >> > & RightLG::getRules() const {
	return rules;
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> >& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightLG::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool RightLG::addRawRule(alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rightHandSide) {
	if(rightHandSide.size() == 0) {
		return addRule(std::move(leftHandSide), std::move(rightHandSide));
	} else if(nonterminalAlphabet.count(rightHandSide[rightHandSide.size() - 1])) {
		return addRule(std::move(leftHandSide), std::make_pair(std::vector<alphabet::Symbol>(std::make_move_iterator(rightHandSide.begin()), std::make_move_iterator(rightHandSide.end()) - 1), std::move(rightHandSide[rightHandSide.size() - 1])));
	} else {
		return addRule(std::move(leftHandSide), std::move(rightHandSide));
	}
}

std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> RightLG::getRawRules() const {
	std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> res;
	for(const auto& rule : getRules()) {
		for(const auto& rhs : rule.second) {
			if(rhs.is<std::vector<alphabet::Symbol>>()) {
				res[rule.first].insert(rhs.get<std::vector<alphabet::Symbol>>());
			} else {
				const auto& rhsTuple = rhs.get<std::pair<std::vector<alphabet::Symbol>, alphabet::Symbol>>();
				std::vector<alphabet::Symbol> tmp;
				tmp.insert(tmp.end(), rhsTuple.first.begin(), rhsTuple.first.end());
				tmp.push_back(rhsTuple.second);
				res[rule.first].insert(tmp);
			}
		}
	}
	return res;
}

bool RightLG::removeRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		return removeRule(leftHandSide, rightHandSide);
	} else if(nonterminalAlphabet.count(rightHandSide[rightHandSide.size() - 1])) {
		return removeRule(leftHandSide, std::make_pair(std::vector<alphabet::Symbol>(rightHandSide.begin(), rightHandSide.end() - 1), rightHandSide[rightHandSide.size() - 1]));
	} else {
		return removeRule(leftHandSide, rightHandSide);
	}
}

int RightLG::compare(const RightLG& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void RightLG::operator>>(std::ostream& out) const {
	out << "(RightLG "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

RightLG::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string RightLG::XML_TAG_NAME = "RightLG";

RightLG RightLG::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RightLG::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	RightLG grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RightLG::XML_TAG_NAME);
	return grammar;
}

void RightLG::parseRule(std::deque<sax::Token>::iterator& input, RightLG& grammar) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS(input);
	std::vector<alphabet::Symbol> rhs = GrammarFromXMLParser::parseRuleRHS(input);
	grammar.addRawRule(std::move(lhs), std::move(rhs));
}

void RightLG::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RightLG::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);

	out.emplace_back(RightLG::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void RightLG::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleSingleSymbolLHS(out, rule.first);
			GrammarToXMLComposer::composeRuleRightLGRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

auto RightLGParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::RightLG>();
auto RightLGParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::RightLG>();

} /* namespace alib */
