/*
 * ContextPreservingUnrestrictedGrammar.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "ContextPreservingUnrestrictedGrammar.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace grammar {

ContextPreservingUnrestrictedGrammar::ContextPreservingUnrestrictedGrammar(alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {

}

ContextPreservingUnrestrictedGrammar::ContextPreservingUnrestrictedGrammar(std::set<alphabet::Symbol> nonterminalAlphabet, std::set<alphabet::Symbol> terminalAlphabet, alphabet::Symbol initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(std::move(initialSymbol)) {
	setNonterminalAlphabet(std::move(nonterminalAlphabet));
	setTerminalAlphabet(std::move(terminalAlphabet));
}

GrammarBase* ContextPreservingUnrestrictedGrammar::clone() const {
	return new ContextPreservingUnrestrictedGrammar(*this);
}

GrammarBase* ContextPreservingUnrestrictedGrammar::plunder() && {
	return new ContextPreservingUnrestrictedGrammar(std::move(*this));
}

bool ContextPreservingUnrestrictedGrammar::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		for(const alphabet::Symbol& lCont : std::get<0>(rule.first))
			if(lCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const alphabet::Symbol& rCont : std::get<2>(rule.first))
			if(rCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool ContextPreservingUnrestrictedGrammar::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		for(const alphabet::Symbol& lCont : std::get<0>(rule.first))
			if(lCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		if(std::get<1>(rule.first) == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const alphabet::Symbol& rCont : std::get<2>(rule.first))
			if(rCont == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool ContextPreservingUnrestrictedGrammar::addRule(std::vector<alphabet::Symbol> lContext, alphabet::Symbol leftHandSide, std::vector<alphabet::Symbol> rContext, std::vector<alphabet::Symbol> rightHandSide) {
	for(const alphabet::Symbol& symbol : lContext) {
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
	}

	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	for(const alphabet::Symbol& symbol : rContext) {
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
	}

	for(const alphabet::Symbol& symbol : rightHandSide) {
		if(!terminalAlphabet.count(symbol) && !nonterminalAlphabet.count(symbol))
			throw GrammarException("Symbol \"" + (std::string) symbol + "\" is not neither terminal nor nonterminal symbol");
	}

	return rules[make_tuple(std::move(lContext), std::move(leftHandSide), std::move(rContext))].insert(std::move(rightHandSide)).second;
}

const std::map<std::tuple<std::vector<alphabet::Symbol>, alphabet::Symbol, std::vector<alphabet::Symbol>>, std::set<std::vector<alphabet::Symbol>>> & ContextPreservingUnrestrictedGrammar::getRules() const {
	return rules;
}

bool ContextPreservingUnrestrictedGrammar::removeRule(const std::vector<alphabet::Symbol>& lContext, const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rContext, const std::vector<alphabet::Symbol>& rightHandSide) {
	return rules[make_tuple(lContext, leftHandSide, rContext)].erase(rightHandSide);
}

int ContextPreservingUnrestrictedGrammar::compare(const ContextPreservingUnrestrictedGrammar& other) const {
	auto first = std::tie(terminalAlphabet, nonterminalAlphabet, initialSymbol, rules);
	auto second = std::tie(other.terminalAlphabet, other.nonterminalAlphabet, other.initialSymbol, other.rules);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void ContextPreservingUnrestrictedGrammar::operator>>(std::ostream& out) const {
	out << "(ContextPreservingUnrestrictedGrammar "
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

ContextPreservingUnrestrictedGrammar::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

const std::string ContextPreservingUnrestrictedGrammar::XML_TAG_NAME = "ContextPreservingUnrestrictedGrammar";

ContextPreservingUnrestrictedGrammar ContextPreservingUnrestrictedGrammar::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, ContextPreservingUnrestrictedGrammar::XML_TAG_NAME);

	std::set<alphabet::Symbol> nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet(input);
	std::set<alphabet::Symbol> terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet(input);
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol(input);

	ContextPreservingUnrestrictedGrammar grammar(std::move(initialSymbol));
	grammar.setNonterminalAlphabet(std::move(nonterminalAlphabet));
	grammar.setTerminalAlphabet(std::move(terminalAlphabet));

	GrammarFromXMLParser::parseRules(input, grammar);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, ContextPreservingUnrestrictedGrammar::XML_TAG_NAME);
	return grammar;
}

void ContextPreservingUnrestrictedGrammar::parseRule(std::deque<sax::Token>::iterator& input, ContextPreservingUnrestrictedGrammar& grammar) {
	std::vector<alphabet::Symbol> lContext = GrammarFromXMLParser::parseRuleLContext(input);
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS(input);
	std::vector<alphabet::Symbol> rContext = GrammarFromXMLParser::parseRuleRContext(input);
	std::vector<alphabet::Symbol> rhs = GrammarFromXMLParser::parseRuleRHS(input);
	grammar.addRule(std::move(lContext), std::move(lhs), std::move(rContext), std::move(rhs));
}

void ContextPreservingUnrestrictedGrammar::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(ContextPreservingUnrestrictedGrammar::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GrammarToXMLComposer::composeNonterminalAlphabet(out, this->getNonterminalAlphabet());
	GrammarToXMLComposer::composeTerminalAlphabet(out, this->getTerminalAlphabet());
	GrammarToXMLComposer::composeInitialSymbol(out, this->getInitialSymbol());
	composeRules(out);

	out.emplace_back(ContextPreservingUnrestrictedGrammar::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

void ContextPreservingUnrestrictedGrammar::composeRules(std::deque<sax::Token>& out) const {
	out.emplace_back("rules", sax::Token::TokenType::START_ELEMENT);
	for (const auto& rule : this->getRules()) {

		for(const auto& rhs : rule.second) {
			out.emplace_back("rule", sax::Token::TokenType::START_ELEMENT);

			GrammarToXMLComposer::composeRuleLContext(out, std::get<0>(rule.first));
			GrammarToXMLComposer::composeRuleSingleSymbolLHS(out, std::get<1>(rule.first));
			GrammarToXMLComposer::composeRuleRContext(out, std::get<2>(rule.first));
			GrammarToXMLComposer::composeRuleRHS(out, rhs);

			out.emplace_back("rule", sax::Token::TokenType::END_ELEMENT);
		}
	}

	out.emplace_back("rules", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

namespace alib {

auto ContextPreservingUnrestrictedGrammarParserRegister = xmlApi<grammar::Grammar>::ParserRegister<grammar::ContextPreservingUnrestrictedGrammar>();
auto ContextPreservingUnrestrictedGrammarParserRegister2 = xmlApi<alib::Object>::ParserRegister<grammar::ContextPreservingUnrestrictedGrammar>();

} /* namespace alib */
