/*
 * GraphFeatures.h
 *
 *  Created on: Jun 19, 2014
 *      Author: David Rosca
 */

#ifndef ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_
#define ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_

#include <set>
#include <vector>
#include <unordered_map>

#include "IUndirectedGraph.h"

namespace graph {

/*
 *  AdjacencyMatrix representation
 *
 *  unordered_map:
 *    [ -- ][ n1 ][ n2 ][ n3 ][ n4 ]
 *    [ n1 ]  0     2     1     1
 *    [ n2 ]  2     0     0     1
 *    [ n3 ]  1     0     0     1
 *    [ n4 ]  1     1     1     0
 *
 *  edges:
 *    n1 <-> n2, n1 <-> n2, n1 <-> n3, n1 <-> n4
 *    n2 <-> n4
 *    n3 <-> n4
 *
 *  undirected edge n1 <-> n2 is represented as
 *  directed edge n1 -> n2 and directed edge n2 -> n1
 *
 */

class AdjacencyMatrixUndirectedGraph : public IUndirectedGraph
{
public:
	std::set<Node> allNodes() const override;
	std::set<UndirectedEdge> allEdges() const override;

	bool hasNode(const Node &node) const override;
	bool hasEdge(const UndirectedEdge &edge) const override;

	bool hasEdge(const Node &first, const Node &second) const override;
	std::set<UndirectedEdge> findEdges(const Node &first, const Node &second) const override;

	std::set<Node> neighbors(const Node &node) const override;
	std::set<UndirectedEdge> neighborEdges(const Node &node) const override;

	bool addNode(const Node &node) override;
	bool removeNode(const Node &node) override;

	bool addEdge(const UndirectedEdge &edge) override;
	bool removeEdge(const UndirectedEdge &edge) override;

	void copy(IUndirectedGraph *other) override;
	int compare(IUndirectedGraph *other) const override;

private:
	std::set<UndirectedEdge> edges;
	std::unordered_map<Node, std::unordered_map<Node, int>> adj;
};

} // namespace graph

#endif // ADJACENCY_MATRIX_UNDIRECTED_GRAPH_H_
