/*
 * GraphFeatures.h
 *
 *  Created on: Jun 19, 2014
 *      Author: David Rosca
 */

#include "UndirectedGraph.h"
#include "IUndirectedGraph.h"
#include "AdjacencyListUndirectedGraph.h"
#include "AdjacencyMatrixUndirectedGraph.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace graph {

UndirectedGraph::UndirectedGraph(REPRESENTATION representation)
	: representation(representation)
{
	init();
}

UndirectedGraph::~UndirectedGraph() noexcept
{
	delete impl;
}

UndirectedGraph::UndirectedGraph(const UndirectedGraph &other)
	: representation(other.representation)
	, nodeValues(other.nodeValues)
	, edgeValues(other.edgeValues)
	, embedding(other.embedding)
{
	init();
	impl->copy(other.impl);
}

UndirectedGraph::UndirectedGraph(UndirectedGraph &&other) noexcept
	: representation(other.representation)
	, impl(other.impl)
	, nodeValues(std::move(other.nodeValues))
	, edgeValues(std::move(other.edgeValues))
	, embedding(std::move(other.embedding))
{
	other.impl = 0;
}

UndirectedGraph &UndirectedGraph::operator=(const UndirectedGraph &other)
{
	if (this == &other) {
		return *this;
	}

	*this = UndirectedGraph(other);
	return *this;
}

UndirectedGraph &UndirectedGraph::operator=(UndirectedGraph &&other) noexcept
{
	this->representation = other.representation;
	std::swap(this->impl, other.impl);
	std::swap(this->nodeValues, other.nodeValues);
	std::swap(this->edgeValues, other.edgeValues);
	std::swap(this->embedding, other.embedding);
	return *this;
}

GraphBase *UndirectedGraph::clone() const
{
	return new UndirectedGraph(*this);
}

GraphBase *UndirectedGraph::plunder() &&
{
	return new UndirectedGraph(std::move(*this));
}

REPRESENTATION UndirectedGraph::getRepresentation() const
{
	return representation;
}

std::set<Node> UndirectedGraph::getNodes() const
{
	return impl->allNodes();
}

std::set<UndirectedEdge> UndirectedGraph::getEdges() const
{
	return impl->allEdges();
}

bool UndirectedGraph::hasNode(const Node &node) const
{
	return impl->hasNode(node);
}

bool UndirectedGraph::hasEdge(const UndirectedEdge &edge) const
{
	return impl->hasEdge(edge);
}

bool UndirectedGraph::hasEdge(const Node &from, const Node &to) const
{
	return impl->hasEdge(from, to);
}

std::set<UndirectedEdge> UndirectedGraph::findEdges(const Node &from, const Node &to) const
{
	return impl->findEdges(from, to);
}

std::set<Node> UndirectedGraph::neighbors(const Node &node) const
{
	return impl->neighbors(node);
}

std::set<UndirectedEdge> UndirectedGraph::neighborEdges(const Node &node) const
{
	return impl->neighborEdges(node);
}

bool UndirectedGraph::addNode(const Node &node)
{
	return impl->addNode(node);
}

bool UndirectedGraph::addNode(const Node &node, int value)
{
	bool success = addNode(node);
	if (success) {
		setNodeValue(node, value);
	}
	return success;
}

bool UndirectedGraph::removeNode(const Node &node)
{
	bool success = impl->removeNode(node);
	if (success) {
		nodeValues.erase(node);
	}
	return success;
}

bool UndirectedGraph::addEdge(const UndirectedEdge &edge)
{
	return impl->addEdge(edge);
}

bool UndirectedGraph::addEdge(const UndirectedEdge &edge, int value)
{
	bool success = addEdge(edge);
	if (success) {
		setEdgeValue(edge, value);
	}
	return success;
}

bool UndirectedGraph::removeEdge(const UndirectedEdge &edge)
{
	bool success = impl->removeEdge(edge);
	if (success) {
		edgeValues.erase(edge);
	}
	return success;
}

int UndirectedGraph::getNodeValue(const Node &node) const
{
	return nodeValues.at(node);
}

void UndirectedGraph::setNodeValue(const Node &node, int value)
{
	nodeValues.insert({node, value});
}

int UndirectedGraph::getEdgeValue(const UndirectedEdge &edge) const
{
	return edgeValues.at(edge);
}

void UndirectedGraph::setEdgeValue(const UndirectedEdge &edge, int value)
{
	edgeValues.insert({edge, value});
}

std::unordered_map < Node, std::vector < Node > > UndirectedGraph::getEmbedding ( ) const
{
	return embedding;
}

void UndirectedGraph::setEmbedding ( const std::unordered_map < Node, std::vector< Node > > embedding )
{
	this->embedding = embedding;
}

void UndirectedGraph::operator>>(std::ostream &out) const
{
	out << "(UndirectedGraph " << static_cast<std::string>(*impl) << ")";
}

int UndirectedGraph::compare(const UndirectedGraph &other) const
{
	if (representation != other.representation) {
		return static_cast<int>(representation) - static_cast<int>(other.representation);
	}

	auto first = std::tie(nodeValues, edgeValues, embedding);
	auto second = std::tie(other.nodeValues, other.edgeValues, other.embedding);

	std::compare<decltype(first)> comp;
	int res = comp(first, second);

	if (res != 0) {
		return res;
	}

	return impl->compare(other.impl);
}

UndirectedGraph::operator std::string() const
{
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

void UndirectedGraph::init()
{
	switch (representation) {
	case REPRESENTATION::ADJACENCY_LIST:
		impl = new AdjacencyListUndirectedGraph();
		break;

	case REPRESENTATION::ADJACENCY_MATRIX:
		impl = new AdjacencyMatrixUndirectedGraph();
		break;

	case REPRESENTATION::INCIDENCE_MATRIX:

	default:
		impl = 0;
	}
}

const std::string UndirectedGraph::XML_TAG_NAME = "UndirectedGraph";

UndirectedGraph UndirectedGraph::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, UndirectedGraph::XML_TAG_NAME);

	UndirectedGraph graph(GraphFromXMLParser::parseRepresentation(input));
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseUndirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseUndirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, UndirectedGraph::XML_TAG_NAME);
	return graph;
}

void UndirectedGraph::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(UndirectedGraph::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeRepresentation(out, representationToString(this->getRepresentation()));
	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(UndirectedGraph::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace graph */

namespace alib {

auto undirectedGraphParserRegister = xmlApi<graph::Graph>::ParserRegister<graph::UndirectedGraph>();
auto undirectedGraphParserRegister2 = xmlApi<alib::Object>::ParserRegister<graph::UndirectedGraph>();

} /* namespace alib */
