/*
 * SuffixTrieFinalMark.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "SuffixTrieFinalMark.h"
#include "SuffixTrieTerminatingSymbol.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/IndexFromXMLParser.h"
#include "../common/IndexToXMLComposer.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace indexes {

SuffixTrieFinalMark::SuffixTrieFinalMark ( std::set < alphabet::Symbol > alphabet ) {
	this->alphabet = std::move ( alphabet );
	this->tree = new SuffixTrieNodeFinalMark ( { }, false );
}

SuffixTrieFinalMark::SuffixTrieFinalMark ( std::set < alphabet::Symbol > alphabet, SuffixTrieNodeFinalMark tree ) {
	this->alphabet = std::move ( alphabet );
	this->tree = NULL;
	setTree ( std::move ( tree ) );
}

SuffixTrieFinalMark::SuffixTrieFinalMark ( SuffixTrieNodeFinalMark tree ) {
	tree.computeMinimalAlphabet ( alphabet );
	this->tree = NULL;
	setTree ( std::move ( tree ) );
}

SuffixTrieFinalMark::SuffixTrieFinalMark ( const SuffixTrieTerminatingSymbol & other ) {
	this->alphabet = other.getAlphabet ( );
	this->alphabet.erase ( other.getTerminatingSymbol ( ) );
	this->tree = NULL;
	setTree ( SuffixTrieNodeFinalMark ( other.getRoot ( ), other.getTerminatingSymbol ( ) ) );
}

SuffixTrieFinalMark::SuffixTrieFinalMark ( const SuffixTrieFinalMark & other ) : tree ( other.tree->clone ( ) ) {
	alphabet = other.alphabet;
	this->tree->attachTree ( this );
}

SuffixTrieFinalMark::SuffixTrieFinalMark ( SuffixTrieFinalMark && other ) noexcept : tree ( other.tree ) {
	alphabet = std::move ( other.alphabet );
	this->tree->attachTree ( this );
	other.tree = NULL;
}

alib::ObjectBase * SuffixTrieFinalMark::clone ( ) const {
	return new SuffixTrieFinalMark ( * this );
}

alib::ObjectBase * SuffixTrieFinalMark::plunder ( ) && {
	return new SuffixTrieFinalMark ( std::move ( * this ) );
}

SuffixTrieFinalMark & SuffixTrieFinalMark::operator =( const SuffixTrieFinalMark & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieFinalMark ( other );

	return * this;
}

SuffixTrieFinalMark & SuffixTrieFinalMark::operator =( SuffixTrieFinalMark && other ) noexcept {
	std::swap ( this->tree, other.tree );
	std::swap ( this->alphabet, other.alphabet );

	return * this;
}

SuffixTrieFinalMark::~SuffixTrieFinalMark ( ) noexcept {
	delete tree;
}

const SuffixTrieNodeFinalMark & SuffixTrieFinalMark::getRoot ( ) const {
	return * tree;
}

SuffixTrieNodeFinalMark & SuffixTrieFinalMark::getRoot ( ) {
	return * tree;
}

void SuffixTrieFinalMark::setTree ( SuffixTrieNodeFinalMark tree ) {
	delete this->tree;
	this->tree = std::move ( tree ).plunder ( );

	if ( !this->tree->attachTree ( this ) ) {
		delete this->tree;
		throw exception::AlibException ( "Input symbols not in the alphabet." );
	}
}

bool SuffixTrieFinalMark::removeSymbolFromAlphabet ( const alphabet::Symbol & symbol ) {
	if ( this->tree->testSymbol ( symbol ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

void SuffixTrieFinalMark::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieFinalMark " << * ( this->tree ) << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieFinalMark & instance ) {
	instance >> out;
	return out;
}

int SuffixTrieFinalMark::compare ( const SuffixTrieFinalMark & other ) const {
	int res = tree->compare ( * other.tree );

	if ( res == 0 ) {
		std::compare < std::set < alphabet::Symbol > > comp;
		res = comp ( alphabet, other.alphabet );
	}

	return res;
}

SuffixTrieFinalMark::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string SuffixTrieFinalMark::XML_TAG_NAME = "SuffixTrieFinalMark";

SuffixTrieFinalMark SuffixTrieFinalMark::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrieFinalMark::XML_TAG_NAME );
	std::set < alphabet::Symbol > rankedAlphabet = IndexFromXMLParser::parseAlphabet ( input );
	SuffixTrieNodeFinalMark * root = IndexFromXMLParser::parseSuffixTrieNodeFinalMark ( input );
	SuffixTrieFinalMark tree ( std::move ( rankedAlphabet ), std::move ( * root ) );

	delete root;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrieFinalMark::XML_TAG_NAME );
	return tree;
}

void SuffixTrieFinalMark::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrieFinalMark::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	IndexToXMLComposer::composeAlphabet ( out, alphabet );
	IndexToXMLComposer::composeNode ( out, * tree );
	out.emplace_back ( SuffixTrieFinalMark::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace indexes */

namespace alib {

auto suffixTreeFinalMarkParserRegister = xmlApi < alib::Object >::ParserRegister < indexes::SuffixTrieFinalMark > ();

} /* namespace alib */
