/*
 * SuffixTrieNodeFinalMark.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "SuffixTrieNodeFinalMark.h"
#include "SuffixTrieNodeTerminatingSymbol.h"
#include "SuffixTrieFinalMark.h"
#include "../../exception/AlibException.h"
#include <sstream>

namespace indexes {

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( std::map < alphabet::Symbol, SuffixTrieNodeFinalMark * > children, bool finalMark ) : children ( std::move ( children ) ), finalMark ( finalMark ), parentTree ( NULL ) {
	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( const SuffixTrieNodeTerminatingSymbol & other, const alphabet::Symbol & terminatingSymbol ) : finalMark ( false ), parentTree ( NULL ) {
	for ( auto & element : other.getChildren ( ) )
		if ( element.first == terminatingSymbol )
			this->finalMark = true;
		else
			this->children.insert ( std::make_pair ( element.first, new SuffixTrieNodeFinalMark ( * element.second, terminatingSymbol ) ) );

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( const SuffixTrieNodeFinalMark & other ) : finalMark ( other.finalMark ), parentTree ( NULL ) {
	for ( const auto & element : other.children )
		children.insert ( std::make_pair ( element.first, element.second->clone ( ) ) );

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark::SuffixTrieNodeFinalMark ( SuffixTrieNodeFinalMark && other ) noexcept : children ( std::move ( other.children ) ), finalMark ( other.finalMark ), parentTree ( NULL ) {
	other.children.clear ( );

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( NULL );
	this->parent = NULL;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::operator =( const SuffixTrieNodeFinalMark & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieNodeFinalMark ( other );

	return * this;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::operator =( SuffixTrieNodeFinalMark && other ) noexcept {
	std::swap ( this->children, other.children );
	std::swap ( this->finalMark, other.finalMark );
	std::swap ( this->parentTree, other.parentTree ); // this->parentTree is stored within other.parentTree and it is reattached on the next line

	for ( auto & element : this->children )
		element.second->parent = this;

	this->attachTree ( other.parentTree );

	return * this;
}

SuffixTrieNodeFinalMark::~SuffixTrieNodeFinalMark ( ) noexcept {
	for ( auto element : children )
		delete element.second;

	children.clear ( );
}

const std::map < const alphabet::Symbol, const SuffixTrieNodeFinalMark * > & SuffixTrieNodeFinalMark::getChildren ( ) const {
	return * reinterpret_cast < const std::map < const alphabet::Symbol, const SuffixTrieNodeFinalMark * > * > ( & children );
}

bool SuffixTrieNodeFinalMark::getFinalMark ( ) const {
	return finalMark;
}

void SuffixTrieNodeFinalMark::setFinalMark ( bool newFinalMark ) {
	finalMark = newFinalMark;
}

const std::map < alphabet::Symbol, SuffixTrieNodeFinalMark * > & SuffixTrieNodeFinalMark::getChildren ( ) {
	return children;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::getChild ( const alphabet::Symbol & symbol ) {
	std::map < alphabet::Symbol, SuffixTrieNodeFinalMark * >::const_iterator iter = children.find ( symbol );

	if ( iter == children.end ( ) ) throw exception::AlibException ( "child does not exist" );

	return * iter->second;
}

const SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::getChild ( const alphabet::Symbol & symbol ) const {
	std::map < alphabet::Symbol, SuffixTrieNodeFinalMark * >::const_iterator iter = children.find ( symbol );

	if ( iter == children.end ( ) ) throw exception::AlibException ( "child does not exist" );

	return * iter->second;
}

bool SuffixTrieNodeFinalMark::hasChild ( const alphabet::Symbol & symbol ) const {
	if ( children.find ( symbol ) == children.end ( ) ) return false;

	return true;
}

SuffixTrieNodeFinalMark & SuffixTrieNodeFinalMark::addChild ( alphabet::Symbol symbol, SuffixTrieNodeFinalMark node ) {
	std::map < alphabet::Symbol, SuffixTrieNodeFinalMark * >::iterator iter = children.find ( symbol );

	if ( iter != children.end ( ) ) throw exception::AlibException ( "child already exist" );

	if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( symbol ) == this->parentTree->getAlphabet ( ).end ( ) ) ) throw exception::AlibException ( "Symbol is not in the alphabet" );

	return * children.insert ( std::make_pair ( std::move ( symbol ), new SuffixTrieNodeFinalMark ( std::move ( node ) ) ) ).first->second;
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::getParent ( ) {
	return parent;
}

const SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::getParent ( ) const {
	return parent;
}

void SuffixTrieNodeFinalMark::swap ( SuffixTrieNodeFinalMark & other ) {
	const alphabet::Alphabet * thisParentTree  = this->parentTree;
	const alphabet::Alphabet * otherParentTree = other.parentTree;

	SuffixTrieNodeFinalMark tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachTree ( thisParentTree );
	other.attachTree ( otherParentTree );
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::clone ( ) const {
	return new SuffixTrieNodeFinalMark ( * this );
}

SuffixTrieNodeFinalMark * SuffixTrieNodeFinalMark::plunder ( ) && {
	return new SuffixTrieNodeFinalMark ( std::move ( * this ) );
}

int SuffixTrieNodeFinalMark::compare ( const SuffixTrieNodeFinalMark & other ) const {
	int thisSize  = this->children.size ( );
	int otherSize = other.children.size ( );

	if ( thisSize < otherSize ) return -1;

	if ( thisSize > otherSize ) return 1;

	auto thisIter  = this->children.begin ( );
	auto otherIter = other.children.begin ( );

	for ( ; thisIter != this->children.end ( ); ++thisIter, ++otherIter ) {
		int res = thisIter->first.compare ( otherIter->first );

		if ( res != 0 ) return res;

		res = thisIter->second->compare ( * otherIter->second );

		if ( res != 0 ) return res;
	}

	return 0;
}

void SuffixTrieNodeFinalMark::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieNodeFinalMark " << " children = " << this->children << " finalMark = " << this->finalMark << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieNodeFinalMark & node ) {
	node >> out;
	return out;
}

bool SuffixTrieNodeFinalMark::testSymbol ( const alphabet::Symbol & symbol ) const {
	for ( const auto & child : this->children ) {
		if ( symbol == child.first ) return true;

		if ( child.second->testSymbol ( symbol ) ) return true;
	}

	return false;
}

bool SuffixTrieNodeFinalMark::attachTree ( const alphabet::Alphabet * tree ) {
	if ( this->parentTree == tree ) return true;

	this->parentTree = tree;

	for ( const auto & child : this->children ) {
		if ( ( this->parentTree != NULL ) && ( this->parentTree->getAlphabet ( ).find ( child.first ) == this->parentTree->getAlphabet ( ).end ( ) ) ) return false;

		if ( !child.second->attachTree ( tree ) ) return false;
	}

	return true;
}

void SuffixTrieNodeFinalMark::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	for ( const auto & child : this->children ) {
		alphabet.insert ( child.first );
		child.second->computeMinimalAlphabet ( alphabet );
	}
}

SuffixTrieNodeFinalMark::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace indexes */
