/*
 * SuffixTrieTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "SuffixTrieTerminatingSymbol.h"
#include "../../exception/AlibException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../sax/FromXMLParserHelper.h"
#include "../common/IndexFromXMLParser.h"
#include "../common/IndexToXMLComposer.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"

namespace indexes {

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol ) : TerminatingSymbolAlphabet ( std::move ( terminatingSymbol ) ) {
	this->alphabet = std::move ( alphabet );
	this->tree = new SuffixTrieNodeTerminatingSymbol ( { } );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : TerminatingSymbolAlphabet ( std::move ( terminatingSymbol ) ) {
	this->alphabet = std::move ( alphabet );
	this->tree = NULL;
	setTree ( std::move ( tree ) );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( alphabet::Symbol terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : TerminatingSymbolAlphabet ( std::move ( terminatingSymbol ) ) {
	tree.computeMinimalAlphabet ( alphabet );
	this->tree = NULL;
	setTree ( std::move ( tree ) );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( const SuffixTrieTerminatingSymbol & other ) : TerminatingSymbolAlphabet ( other.terminatingSymbol ), tree ( other.tree->clone ( ) ) {
	alphabet = other.alphabet;
	this->tree->attachTree ( this );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( SuffixTrieTerminatingSymbol && other ) noexcept : TerminatingSymbolAlphabet ( std::move ( other.terminatingSymbol ) ), tree ( other.tree ) {
	alphabet = std::move ( other.alphabet );
	this->tree->attachTree ( this );
	other.tree = NULL;
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::clone ( ) const {
	return new SuffixTrieTerminatingSymbol ( * this );
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::plunder ( ) && {
	return new SuffixTrieTerminatingSymbol ( std::move ( * this ) );
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( const SuffixTrieTerminatingSymbol & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieTerminatingSymbol ( other );

	return * this;
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( SuffixTrieTerminatingSymbol && other ) noexcept {
	std::swap ( this->tree, other.tree );
	std::swap ( this->alphabet, other.alphabet );

	return * this;
}

SuffixTrieTerminatingSymbol::~SuffixTrieTerminatingSymbol ( ) noexcept {
	delete tree;
}

const SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) const {
	return * tree;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) {
	return * tree;
}

void SuffixTrieTerminatingSymbol::setTree ( SuffixTrieNodeTerminatingSymbol tree ) {
	delete this->tree;
	this->tree = std::move ( tree ).plunder ( );

	if ( !this->tree->attachTree ( this ) ) {
		delete this->tree;
		throw exception::AlibException ( "Input symbols not in the alphabet." );
	}
}

bool SuffixTrieTerminatingSymbol::removeSymbolFromAlphabet ( const alphabet::Symbol & symbol ) {
	if ( this->tree->testSymbol ( symbol ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

void SuffixTrieTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieTerminatingSymbol " << * ( this->tree ) << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieTerminatingSymbol & instance ) {
	instance >> out;
	return out;
}

int SuffixTrieTerminatingSymbol::compare ( const SuffixTrieTerminatingSymbol & other ) const {
	int res = tree->compare ( * other.tree );

	if ( res == 0 ) {
		std::compare < std::set < alphabet::Symbol > > comp;
		res = comp ( alphabet, other.alphabet );
	}

	return res;
}

SuffixTrieTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

const std::string SuffixTrieTerminatingSymbol::XML_TAG_NAME = "SuffixTrieTerminatingSymbol";

SuffixTrieTerminatingSymbol SuffixTrieTerminatingSymbol::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrieTerminatingSymbol::XML_TAG_NAME );
	std::set < alphabet::Symbol > rankedAlphabet = IndexFromXMLParser::parseAlphabet ( input );
	alphabet::Symbol terminatingSymbol = alib::xmlApi < alphabet::Symbol >::parse ( input );
	SuffixTrieNodeTerminatingSymbol * root = IndexFromXMLParser::parseSuffixTrieNodeTerminatingSymbol ( input );
	SuffixTrieTerminatingSymbol tree ( std::move ( rankedAlphabet ), terminatingSymbol, std::move ( * root ) );

	delete root;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrieTerminatingSymbol::XML_TAG_NAME );
	return tree;
}

void SuffixTrieTerminatingSymbol::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrieTerminatingSymbol::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	IndexToXMLComposer::composeAlphabet ( out, alphabet );
	alib::xmlApi < alphabet::Symbol >::compose ( out, terminatingSymbol );
	IndexToXMLComposer::composeNode ( out, * tree );
	out.emplace_back ( SuffixTrieTerminatingSymbol::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace indexes */

namespace alib {

auto suffixTreeTerminatingSymbolParserRegister = xmlApi < alib::Object >::ParserRegister < indexes::SuffixTrieTerminatingSymbol > ();

} /* namespace alib */
