/*
 * FormalRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_ELEMENT_H_
#define FORMAL_REG_EXP_ELEMENT_H_

#include "../../XmlApi.hpp"
#include "../../common/visitor.hpp"
#include "../../alphabet/Symbol.h"
#include "../common/RegExpAlphabet.h"
#include <set>

namespace regexp {

class FormalRegExp;

class FormalRegExpAlternation;
class FormalRegExpConcatenation;
class FormalRegExpIteration;
class FormalRegExpSymbol;
class FormalRegExpEmpty;
class FormalRegExpEpsilon;

class UnboundedRegExpElement;
class FormalRegExpElement;

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
typedef std::acceptor_base < FormalRegExpElement, FormalRegExpAlternation, FormalRegExpConcatenation, FormalRegExpIteration, FormalRegExpSymbol, FormalRegExpEmpty, FormalRegExpEpsilon > VisitableFormalRegExpElement;

class FormalRegExpElement : public alib::base < FormalRegExpElement >, public VisitableFormalRegExpElement {
protected:
	/*
	 * Parent regexp contanining this instance of RegExpElement
	 */
	const RegExpAlphabet * parentRegExp;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const RegExpAlphabet * regexp ) = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const = 0;

public:
	explicit FormalRegExpElement ( );

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual UnboundedRegExpElement * cloneAsUnbounded ( ) const = 0;

	friend class FormalRegExp;

	friend class FormalRegExpAlternation;
	friend class FormalRegExpConcatenation;
	friend class FormalRegExpIteration;
	friend class FromalRegExpSymbol;
	friend class FormalRegExpEmpty;
	friend class FormalRegExpEpsilon;
};

} /* namespace regexp */

namespace alib {

template < >
struct xmlApi < regexp::FormalRegExpElement * > {
	static regexp::FormalRegExpElement * parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::FormalRegExpElement * const & data );
};

template < >
struct xmlApi < const regexp::FormalRegExpElement * > {
	static void compose ( std::deque < sax::Token > & output, const regexp::FormalRegExpElement * const & data );
};

} /* namespace alib */

#endif /* FORMAL_REG_EXP_ELEMENT_H_ */
