/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_ELEMENT_H_
#define UNBOUNDED_REG_EXP_ELEMENT_H_

#include "../../XmlApi.hpp"
#include "../../common/visitor.hpp"
#include "../../alphabet/Symbol.h"
#include "../common/RegExpAlphabet.h"
#include <set>

namespace regexp {

class UnboundedRegExp;

class UnboundedRegExpAlternation;
class UnboundedRegExpConcatenation;
class UnboundedRegExpIteration;
class UnboundedRegExpSymbol;
class UnboundedRegExpEmpty;
class UnboundedRegExpEpsilon;

class FormalRegExpElement;
class UnboundedRegExpElement;

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
typedef std::acceptor_base < UnboundedRegExpElement, UnboundedRegExpAlternation, UnboundedRegExpConcatenation, UnboundedRegExpIteration, UnboundedRegExpSymbol, UnboundedRegExpEmpty, UnboundedRegExpEpsilon > VisitableUnboundedRegExpElement;

class UnboundedRegExpElement : public alib::base < UnboundedRegExpElement >, public VisitableUnboundedRegExpElement {
protected:
	/**
	 * Parent regexp contanining this instance of RegExpElement
	 */
	const RegExpAlphabet * parentRegExp;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const RegExpAlphabet * regexp ) = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const = 0;

public:
	explicit UnboundedRegExpElement ( );

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual FormalRegExpElement * cloneAsFormal ( ) const = 0;

	friend class UnboundedRegExp;

	friend class UnboundedRegExpAlternation;
	friend class UnboundedRegExpConcatenation;
	friend class UnboundedRegExpIteration;
	friend class FromalRegExpSymbol;
	friend class UnboundedRegExpEmpty;
	friend class UnboundedRegExpEpsilon;
};

} /* namespace regexp */

namespace alib {

template < >
struct xmlApi < regexp::UnboundedRegExpElement * > {
	static regexp::UnboundedRegExpElement * parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpElement * const & data );
};

template < >
struct xmlApi < const regexp::UnboundedRegExpElement * > {
	static void compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpElement * const & data );
};

} /* namespace alib */

#endif /* UNBOUNDED_REG_EXP_ELEMENT_H_ */
