/*
 * CyclicString.cpp
 *
 *  Created on: Nov 23, 2013
 *	  Author: Jan Travnicek
 */

#include "CyclicString.h"
#include "Epsilon.h"
#include "../exception/AlibException.h"

#include <sstream>
#include <algorithm>

#include "../sax/FromXMLParserHelper.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "String.h"
#include "../object/Object.h"
#include "../XmlApi.hpp"
#include "../CastApi.hpp"

namespace string {

CyclicString::CyclicString() {

}

CyclicString::CyclicString(std::set<alphabet::Symbol> alphabet, std::vector<alphabet::Symbol> data) {
	this->alphabet = std::move(alphabet);
	setContent(std::move(data));
}

CyclicString::CyclicString(std::vector<alphabet::Symbol> data) {
	alphabet = std::set<alphabet::Symbol>(data.begin(), data.end());
	setContent(std::move(data));
}

CyclicString::CyclicString(const std::string& string) {
	for(const char& symbol : string) {
		m_Data.push_back( alphabet::symbolFrom( symbol ) );
	}
	alphabet = std::set<alphabet::Symbol>(m_Data.begin(), m_Data.end());
}

CyclicString::CyclicString(const char* string) : CyclicString((std::string) string) {
}

CyclicString::CyclicString(const Epsilon& epsilon) {
	alphabet = epsilon.getAlphabet();
}

StringBase* CyclicString::clone() const {
	return new CyclicString(*this);
}

StringBase* CyclicString::plunder() && {
	return new CyclicString(std::move(*this));
}

bool CyclicString::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(std::any_of(m_Data.begin(), m_Data.end(), [&](const alphabet::Symbol & s) { return s == symbol; } ) )
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");

	return alphabet.erase(symbol);
}

const std::vector<alphabet::Symbol>& CyclicString::getContent() const {
	return this->m_Data;
}

//serves as both move and copy content setter
void CyclicString::setContent(std::vector<alphabet::Symbol> data) {
	std::set<alphabet::Symbol> minimalAlphabet(data.begin(), data.end());
	std::set<alphabet::Symbol> unknownSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), alphabet.begin(), alphabet.end(), std::inserter(unknownSymbols, unknownSymbols.end()));

	if(unknownSymbols.size() > 0)
		throw exception::AlibException("Input symbols not in the alphabet.");

	m_Data = std::move(data);
}

bool CyclicString::isEmpty() const {
	return this->m_Data.size() == 0;
}

int CyclicString::compare(const CyclicString& other) const {
	auto first = std::tie(m_Data, alphabet);
	auto second = std::tie(other.m_Data, other.alphabet);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void CyclicString::operator >>(std::ostream& out) const {
	out << "(CyclicString ";
	for(const alphabet::Symbol& symbol : this->m_Data)
		out << symbol;
	out << ")";
}

CyclicString::operator std::string () const {
	std::stringstream ss;
	ss << "<";
	for(const alphabet::Symbol& symbol : this->m_Data)
		ss << (std::string) symbol;
	ss << ">";
	return std::move(ss).str();
}

const std::string CyclicString::XML_TAG_NAME = "CyclicString";

CyclicString CyclicString::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, CyclicString::XML_TAG_NAME);
	std::set<alphabet::Symbol> alphabet = StringFromXMLParser::parseAlphabet(input);
	std::vector<alphabet::Symbol> content = StringFromXMLParser::parseContent(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, CyclicString::XML_TAG_NAME);
	return CyclicString(alphabet, content);
}

void CyclicString::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(CyclicString::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT);
	StringToXMLComposer::compose(out, alphabet);
	StringToXMLComposer::compose(out, m_Data);
	out.emplace_back(CyclicString::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT);
}

} /* namespace string */

namespace alib {

auto cyclicStringParserRegister = xmlApi<string::String>::ParserRegister<string::CyclicString>();
auto cyclicStringParserRegister2 = xmlApi<alib::Object>::ParserRegister<string::CyclicString>();

auto CyclicStringFromEpsilon = castApi::CastRegister<string::CyclicString, string::Epsilon>();
auto CyclicStringCastBinder = castApi::CastPoolStringBinder<string::CyclicString>(string::CyclicString::XML_TAG_NAME);

} /* namespace alib */
