/*
 * PrefixRankedBarTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixRankedBarTree.h"
#include "../../exception/AlibException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "RankedTree.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"
#include "../../CastApi.hpp"

#include "../../alphabet/BarSymbol.h"

namespace tree {

PrefixRankedBarTree::PrefixRankedBarTree ( alphabet::Symbol bar, std::set < alphabet::RankedSymbol > alphabet, std::vector < alphabet::RankedSymbol > data ) : bar ( std::move ( bar ) ) {
	addSymbolsToAlphabet ( std::move ( alphabet ) );
	setContent ( std::move ( data ) );
}

PrefixRankedBarTree::PrefixRankedBarTree ( alphabet::Symbol bar, std::vector < alphabet::RankedSymbol > data ) : bar ( std::move ( bar ) ) {
	arityChecksum ( data );

	addSymbolsToAlphabet ( std::set < alphabet::RankedSymbol > ( data.begin ( ), data.end ( ) ) );
	m_Data = std::move ( data );
}

PrefixRankedBarTree::PrefixRankedBarTree ( alphabet::Symbol bar, const RankedTree & tree ) : bar ( std::move ( bar ) ) {
	toPrefixRankedBar ( tree.getRoot ( ) );

	for ( const alphabet::RankedSymbol & symbol : tree.getAlphabet ( ) ) {
		addSymbolToAlphabet ( symbol );
		addSymbolToAlphabet ( alphabet::RankedSymbol ( alphabet::Symbol ( alphabet::BarSymbol::BAR ), symbol.getRank ( ) ) );
	}
}

PrefixRankedBarTree::PrefixRankedBarTree ( const RankedTree & tree ) : PrefixRankedBarTree ( alphabet::Symbol ( alphabet::BarSymbol::BAR ), tree ) {
}

void PrefixRankedBarTree::toPrefixRankedBar ( const RankedNode & node ) {
	m_Data.push_back ( node.getSymbol ( ) );

	for ( const RankedNode * child : node.getChildren ( ) )
		toPrefixRankedBar ( * child );

	m_Data.push_back ( alphabet::RankedSymbol ( bar, node.getSymbol ( ).getRank ( ) ) );
}

RankedTreeBase * PrefixRankedBarTree::clone ( ) const {
	return new PrefixRankedBarTree ( * this );
}

RankedTreeBase * PrefixRankedBarTree::plunder ( ) && {
	return new PrefixRankedBarTree ( std::move ( * this ) );
}

bool PrefixRankedBarTree::removeSymbolFromAlphabet ( const alphabet::RankedSymbol & symbol ) {
	if ( std::any_of ( m_Data.begin ( ), m_Data.end ( ), [&] ( const alphabet::RankedSymbol & s ) {
			return s == symbol;
		} ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

const std::vector < alphabet::RankedSymbol > & PrefixRankedBarTree::getContent ( ) const {
	return this->m_Data;
}

const alphabet::Symbol & PrefixRankedBarTree::getBarSymbol ( ) const {
	return bar;
}

void PrefixRankedBarTree::setContent ( std::vector < alphabet::RankedSymbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::RankedSymbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::RankedSymbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), alphabet.begin ( ), alphabet.end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::AlibException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixRankedBarTree::arityChecksum ( const std::vector < alphabet::RankedSymbol > & data ) {
	int arityChecksumTerminals = 1;
	int arityChecksumBars  = 1;
	int arityChecksumTypes = 0;

	for ( const alphabet::RankedSymbol & symbol : data ) {
		if ( symbol.getSymbol ( ) != bar ) {
			arityChecksumTerminals += symbol.getRank ( ).getData ( );
			arityChecksumTerminals -= 1;
			arityChecksumTypes += 1;
		} else {
			arityChecksumBars  += symbol.getRank ( ).getData ( );
			arityChecksumBars  -= 1;
			arityChecksumTypes -= 1;
		}
	}

	if ( ( arityChecksumTerminals != 0 ) || ( arityChecksumBars != 0 ) || ( arityChecksumTypes != 0 ) ) throw exception::AlibException ( "The string does not form a tree" );
}

bool PrefixRankedBarTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixRankedBarTree::compare ( const PrefixRankedBarTree & other ) const {
	auto first	= std::tie ( m_Data, alphabet );
	auto second = std::tie ( other.m_Data, other.alphabet );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixRankedBarTree::operator >>( std::ostream & out ) const {
	if ( this->isEmpty ( ) ) {
		out << "(Epsilon)";
	} else {
		out << "(PrefixRankedBarTree ";

		for ( const alphabet::RankedSymbol & symbol : this->m_Data )
			out << symbol;

		out << ")";
	}
}

PrefixRankedBarTree::operator std::string ( ) const {
	std::stringstream ss;

	if ( this->isEmpty ( ) ) {
		ss << "E";
	} else {
		ss << "\"";

		for ( const alphabet::RankedSymbol & symbol : this->m_Data )
			ss << ( std::string ) symbol;

		ss << "\"";
	}

	return std::move ( ss ).str ( );
}

const std::string PrefixRankedBarTree::XML_TAG_NAME = "PrefixRankedBarTree";

PrefixRankedBarTree PrefixRankedBarTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedBarTree::XML_TAG_NAME );
	alphabet::Symbol bar = TreeFromXMLParser::parseBar ( input );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	std::vector < alphabet::RankedSymbol > data = TreeFromXMLParser::parseRankedContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedBarTree::XML_TAG_NAME );

	return PrefixRankedBarTree ( std::move ( bar ), std::move ( rankedAlphabet ), std::move ( data ) );
}

void PrefixRankedBarTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedBarTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBar ( out, bar );
	TreeToXMLComposer::composeAlphabet ( out, alphabet );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedBarTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto prefixRankedBarTreeParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::PrefixRankedBarTree > ();
auto prefixRankedBarTreeParserRegister2 = xmlApi < tree::RankedTreeWrapper >::ParserRegister < tree::PrefixRankedBarTree > ();
auto prefixRankedBarTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::PrefixRankedBarTree > ();

auto PrefixRankedBarTreeFromRankedTree = castApi::CastRegister < tree::PrefixRankedBarTree, tree::RankedTree > ( );
auto PrefixRankedBarTreeCastBinder = castApi::CastPoolStringBinder < tree::PrefixRankedBarTree > ( tree::PrefixRankedBarTree::XML_TAG_NAME );

} /* namespace alib */
