/*
 * RankedNode.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_NODE_H_
#define RANKED_NODE_H_

#include "../../alphabet/LabeledSymbol.h"
#include "../../alphabet/RankedSymbol.h"
#include <vector>
#include <set>
#include "../common/RankedAlphabet.h"

namespace tree {

class UnrankedNode;

/**
 * Represents a node in the ranked tree. Contains name of the symbol.
 */
class RankedNode {
protected:
	alphabet::RankedSymbol symbol;

	std::vector < RankedNode * > children;

	RankedNode * parent;

	/**
	 * Parent tree contanining this instance of RankedTree
	 */
	const RankedAlphabet * parentTree;

	/**
	 * @copydoc RankedNode::cloneAsUnranked() const
	 */
	UnrankedNode * cloneAsUnranked ( ) const;

	/**
	 * @copydoc RankedNode::testSymbol() const
	 */
	bool testSymbol ( const alphabet::RankedSymbol & symbol ) const;

	/**
	 * @copydoc RankedNode::attachTree()
	 */
	bool attachTree ( const RankedAlphabet * tree );

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < alphabet::RankedSymbol > & alphabet ) const;

public:
	explicit RankedNode ( alphabet::RankedSymbol symbol, std::vector < RankedNode * > children );

	RankedNode ( const RankedNode & other );
	RankedNode ( RankedNode && other ) noexcept;
	RankedNode & operator =( const RankedNode & other );
	RankedNode & operator =( RankedNode && other ) noexcept;
	~RankedNode ( ) noexcept;

	/**
	 * @copydoc RankedNode::clone() const
	 */
	RankedNode * clone ( ) const;

	/**
	 * @copydoc RankedNode::plunder() const
	 */
	RankedNode * plunder ( ) &&;

	/**
	 * @return children
	 */
	const std::vector < const RankedNode * > & getChildren ( ) const;

	/**
	 * @return children
	 */
	const std::vector < RankedNode * > & getChildren ( );

	RankedNode * getParent ( );

	const RankedNode * getParent ( ) const;

	void setSymbol ( alphabet::RankedSymbol symbol );

	void swap ( RankedNode & other );

	int compare ( const RankedNode & ) const;

	/**
	 * @copydoc RankedNode::operator>>() const
	 */
	void operator >>( std::ostream & out ) const;

	friend std::ostream & operator <<( std::ostream &, const RankedNode & node );

	/**
	 * Returns the string representation of Node's Symbol.
	 */
	const alphabet::RankedSymbol & getSymbol ( ) const;

	explicit operator std::string ( ) const;

	void nicePrint ( std::ostream &, const std::string & = "", const bool = true ) const;

	friend class UnrankedTree;
	friend class UnrankedPattern;
	friend class RankedTree;
	friend class RankedPattern;
};

} /* namespace tree */

namespace alib {

template < >
struct xmlApi < tree::RankedNode * > {
	static tree::RankedNode * parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const tree::RankedNode * const & data );
};

template < >
struct xmlApi < const tree::RankedNode * > {
	static void compose ( std::deque < sax::Token > & output, const tree::RankedNode * const & data );
};

} /* namespace alib */

#endif /* RANKED_NODE_H_ */
