/*
 * RankedPattern.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef RANKED_PATTERN_H_
#define RANKED_PATTERN_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "RankedNode.h"
#include "../RankedTreeBase.h"
#include "../common/RankedPatternAlphabet.h"

namespace tree {

class RankedNode;
class UnrankedPattern;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a pattern of RegExpElement.
 */
class RankedPattern : public RankedTreeBase, public RankedPatternAlphabet {
protected:
	RankedNode * pattern;

public:
	/**
	 * @copydoc RankedNode::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc RankedNode::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedPattern ( const UnrankedPattern & other );
	explicit RankedPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > alphabet, RankedNode pattern );
	explicit RankedPattern ( alphabet::RankedSymbol subtreeWildcard, RankedNode pattern );

	/**
	 * Copy constructor.
	 * @param other pattern to copy
	 */
	RankedPattern ( const RankedPattern & other );
	RankedPattern ( RankedPattern && other ) noexcept;
	RankedPattern & operator =( const RankedPattern & other );
	RankedPattern & operator =( RankedPattern && other ) noexcept;
	~RankedPattern ( ) noexcept;

	/**
	 * @return Root node of the regular expression pattern
	 */
	const RankedNode & getRoot ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return RankedPatternAlphabet::getAlphabet ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		addSymbolsToAlphabet ( symbols );
	}

	/**
	 * @return Root node of the regular expression pattern
	 */
	RankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression pattern
	 * @param pattern root node to set
	 */
	void setTree ( RankedNode pattern );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const alphabet::RankedSymbol & symbol );

	/**
	 * Prints XML representation of the pattern to the output stream.
	 * @param out output stream to which print the pattern
	 * @param pattern pattern to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedPattern & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static RankedPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
struct compare < tree::RankedPattern > {
	int operator ()( const tree::RankedPattern & first, const tree::RankedPattern & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* RANKED_PATTERN_H_ */
