/*
 * RankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef RANKED_TREE_H_
#define RANKED_TREE_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "RankedNode.h"
#include "../RankedTreeBase.h"
#include "../common/RankedAlphabet.h"

namespace tree {

class RankedNode;
class UnrankedTree;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class RankedTree : public RankedTreeBase, public RankedAlphabet {
protected:
	RankedNode * tree;

public:
	/**
	 * @copydoc RankedNode::clone() const
	 */
	virtual RankedTreeBase * clone ( ) const;

	/**
	 * @copydoc RankedNode::plunder() const
	 */
	virtual RankedTreeBase * plunder ( ) &&;

	explicit RankedTree ( const UnrankedTree & other );
	explicit RankedTree ( std::set < alphabet::RankedSymbol > alphabet, RankedNode tree );
	explicit RankedTree ( RankedNode tree );

	/**
	 * Copy constructor.
	 * @param other tree to copy
	 */
	RankedTree ( const RankedTree & other );
	RankedTree ( RankedTree && other ) noexcept;
	RankedTree & operator =( const RankedTree & other );
	RankedTree & operator =( RankedTree && other ) noexcept;
	~RankedTree ( ) noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const RankedNode & getRoot ( ) const;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return RankedAlphabet::getAlphabet ( );
	}

	void extendAlphabet ( const std::set < alphabet::RankedSymbol > & symbols ) {
		addSymbolsToAlphabet ( symbols );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	RankedNode & getRoot ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( RankedNode tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const alphabet::RankedSymbol & symbol );

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RankedTree & other ) const;

	void nicePrint ( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	const static std::string XML_TAG_NAME;

	static RankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace tree */

namespace std {

template < >
struct compare < tree::RankedTree > {
	int operator ()( const tree::RankedTree & first, const tree::RankedTree & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* RANKED_TREE_H_ */
