/*
 * PrefixBarTree.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "PrefixBarTree.h"
#include "../../exception/AlibException.h"

#include <sstream>
#include <algorithm>
#include <deque>

#include "UnrankedTree.h"

#include "../../sax/FromXMLParserHelper.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../UnrankedTreeWrapper.h"
#include "../../object/Object.h"
#include "../../XmlApi.hpp"
#include "../../CastApi.hpp"

#include "../../alphabet/BarSymbol.h"

namespace tree {

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::set < alphabet::Symbol > alphabet, std::vector < alphabet::Symbol > data ) : bar ( std::move ( bar ) ) {
	addSymbolsToAlphabet ( std::move ( alphabet ) );
	setContent ( std::move ( data ) );
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, std::vector < alphabet::Symbol > data ) : bar ( std::move ( bar ) ) {
	arityChecksum ( data );

	addSymbolsToAlphabet ( std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) ) );
	m_Data = std::move ( data );
}

PrefixBarTree::PrefixBarTree ( alphabet::Symbol bar, const UnrankedTree & tree ) : bar ( std::move ( bar ) ) {
	toPrefixBar ( tree.getRoot ( ) );

	addSymbolsToAlphabet ( tree.getAlphabet ( ) );
	addSymbolToAlphabet ( alphabet::Symbol ( alphabet::BarSymbol::BAR ) );
}

PrefixBarTree::PrefixBarTree ( const UnrankedTree & tree ) : PrefixBarTree ( alphabet::Symbol ( alphabet::BarSymbol::BAR ), tree ) {
}

void PrefixBarTree::toPrefixBar ( const UnrankedNode & node ) {
	m_Data.push_back ( node.getSymbol ( ) );

	for ( const UnrankedNode * child : node.getChildren ( ) )
		toPrefixBar ( * child );

	m_Data.push_back ( bar );
}

UnrankedTreeBase * PrefixBarTree::clone ( ) const {
	return new PrefixBarTree ( * this );
}

UnrankedTreeBase * PrefixBarTree::plunder ( ) && {
	return new PrefixBarTree ( std::move ( * this ) );
}

bool PrefixBarTree::removeSymbolFromAlphabet ( const alphabet::Symbol & symbol ) {
	if ( std::any_of ( m_Data.begin ( ), m_Data.end ( ), [&] ( const alphabet::Symbol & s ) {
			return s == symbol;
		} ) )
		throw exception::AlibException ( "Input symbol \"" + ( std::string ) symbol + "\" is used." );

	return alphabet.erase ( symbol );
}

const std::vector < alphabet::Symbol > & PrefixBarTree::getContent ( ) const {
	return this->m_Data;
}

const alphabet::Symbol & PrefixBarTree::getBarSymbol ( ) const {
	return bar;
}

void PrefixBarTree::setContent ( std::vector < alphabet::Symbol > data ) {
	arityChecksum ( data );

	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), alphabet.begin ( ), alphabet.end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::AlibException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

void PrefixBarTree::arityChecksum ( const std::vector < alphabet::Symbol > & data ) {
	int arityChecksumTypes = 0;

	for ( const alphabet::Symbol & symbol : data ) {
		if ( symbol != bar )
			arityChecksumTypes += 1;
		else
			arityChecksumTypes -= 1;
	}

	if ( arityChecksumTypes != 0 ) throw exception::AlibException ( "The string does not form a tree" );
}

bool PrefixBarTree::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int PrefixBarTree::compare ( const PrefixBarTree & other ) const {
	auto first	= std::tie ( m_Data, alphabet );
	auto second = std::tie ( other.m_Data, other.alphabet );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void PrefixBarTree::operator >>( std::ostream & out ) const {
	if ( this->isEmpty ( ) ) {
		out << "(Epsilon)";
	} else {
		out << "(PrefixBarTree ";

		for ( const alphabet::Symbol & symbol : this->m_Data )
			out << symbol;

		out << ")";
	}
}

PrefixBarTree::operator std::string ( ) const {
	std::stringstream ss;

	if ( this->isEmpty ( ) ) {
		ss << "E";
	} else {
		ss << "\"";

		for ( const alphabet::Symbol & symbol : this->m_Data )
			ss << ( std::string ) symbol;

		ss << "\"";
	}

	return std::move ( ss ).str ( );
}

const std::string PrefixBarTree::XML_TAG_NAME = "PrefixBarTree";

PrefixBarTree PrefixBarTree::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixBarTree::XML_TAG_NAME );
	alphabet::Symbol bar = TreeFromXMLParser::parseBar ( input );
	std::set < alphabet::Symbol > unrankedAlphabet = TreeFromXMLParser::parseUnrankedAlphabet ( input );
	std::vector < alphabet::Symbol > data = TreeFromXMLParser::parseUnrankedContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixBarTree::XML_TAG_NAME );

	return PrefixBarTree ( std::move ( bar ), std::move ( unrankedAlphabet ), std::move ( data ) );
}

void PrefixBarTree::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixBarTree::XML_TAG_NAME, sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBar ( out, bar );
	TreeToXMLComposer::composeAlphabet ( out, alphabet );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixBarTree::XML_TAG_NAME, sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace alib {

auto prefixBarTreeParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::PrefixBarTree > ();
auto prefixBarTreeParserRegister2 = xmlApi < tree::UnrankedTreeWrapper >::ParserRegister < tree::PrefixBarTree > ();
auto prefixBarTreeParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::PrefixBarTree > ();

auto PrefixBarTreeFromUnrankedTree = castApi::CastRegister < tree::PrefixBarTree, tree::UnrankedTree > ( );
auto PrefixBarTreeCastBinder = castApi::CastPoolStringBinder < tree::PrefixBarTree > ( tree::PrefixBarTree::XML_TAG_NAME );

} /* namespace alib */
