#include "VisitorTest.h"
#include "common/visitor.hpp"
#include "common/base.hpp"
#include <set>
#include <string>

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION ( VisitorTest, "std" );
CPPUNIT_TEST_SUITE_REGISTRATION ( VisitorTest );

void VisitorTest::setUp ( ) {
}

void VisitorTest::tearDown ( ) {
}

namespace visitor {

class Tmp1;
class Tmp2;
class Tmp3;

class TmpBase;

typedef std::acceptor_base < TmpBase, Tmp1, Tmp2, Tmp3 > VisitableTmpBase;

class TmpBase : public alib::base < TmpBase >, public VisitableTmpBase {
};

class Tmp1 : public std::acceptor < Tmp1, VisitableTmpBase, TmpBase > {
	int data;

public:
	Tmp1 ( int data ) : data ( data ) {
	}

	TmpBase * clone ( ) const {
		return new Tmp1 ( * this );
	}

	TmpBase * plunder ( ) && {
		return new Tmp1 ( * this );
	}

	virtual int compare ( const TmpBase & other ) const {
		return -other.compare ( * this );
	}

	virtual int compare ( const Tmp1 & other ) const {
		return this->data - other.data;
	}

	virtual void operator >>( std::ostream & os ) const {
		os << "Tmp1(" << data << ")";
	}

	virtual operator std::string ( ) const {
		return "Tmp1(" + std::to_string ( data ) + ")";
	}

	int getData ( ) const {
		return data;
	}

};

class Tmp2 : public std::acceptor < Tmp2, VisitableTmpBase, TmpBase > {
	double data;

public:
	Tmp2 ( double data ) : data ( data ) {
	}

	Tmp2 ( const Tmp1 & other ) : data ( other.getData ( ) ) {
	}

	TmpBase * clone ( ) const {
		return new Tmp2 ( * this );
	}

	TmpBase * plunder ( ) && {
		return new Tmp2 ( * this );
	}

	virtual int compare ( const TmpBase & other ) const {
		return -other.compare ( * this );
	}

	virtual int compare ( const Tmp2 & other ) const {
		return this->data - other.data;
	}

	virtual void operator >>( std::ostream & os ) const {
		os << "Tmp2(" << data << ")";
	}

	virtual operator std::string ( ) const {
		return "Tmp2(" + std::to_string ( data ) + ")";
	}

	double getData ( ) const {
		return data;
	}

};

class Tmp3 : public std::acceptor < Tmp3, VisitableTmpBase, TmpBase > {
	std::string data;

public:
	Tmp3 ( const std::string & data ) : data ( data ) {
	}

	Tmp3 ( const Tmp1 & other ) : data ( std::to_string ( other.getData ( ) ) ) {
	}

	Tmp3 ( const Tmp2 & other ) : data ( std::to_string ( other.getData ( ) ) ) {
	}

	TmpBase * clone ( ) const {
		return new Tmp3 ( * this );
	}

	TmpBase * plunder ( ) && {
		return new Tmp3 ( * this );
	}

	virtual int compare ( const TmpBase & other ) const {
		return -other.compare ( * this );
	}

	virtual int compare ( const Tmp3 & other ) const {
		return this->data.compare ( other.data );
	}

	virtual void operator >>( std::ostream & os ) const {
		os << "Tmp3(" << data << ")";
	}

	virtual operator std::string ( ) const {
		return "Tmp3(" + data + ")";
	}

	const std::string & getData ( ) const {
		return data;
	}

};

class TmpVisitor : public VisitableTmpBase::const_visitor_type {
	void Visit ( void * userData, const Tmp1 & first ) const {
		int & data = * ( ( int * ) userData );

		data = 1;
		std::cout << first << std::endl;
	}

	void Visit ( void * userData, const Tmp2 & first ) const {
		int & data = * ( ( int * ) userData );

		data = 2;
		std::cout << first << std::endl;
	}

	void Visit ( void * userData, const Tmp3 & first ) const {
		int & data = * ( ( int * ) userData );

		data = 3;
		std::cout << first << std::endl;
	}

};

class TmpSameVisitor : public VisitableTmpBase::const_same_visitor_type {
	void Visit ( void * userData, const Tmp1 & first, const Tmp1 & second ) const {
		int & data = * ( ( int * ) userData );

		data = 1;
		std::cout << first << " " << second << std::endl;
	}

	void Visit ( void * userData, const Tmp2 & first, const Tmp2 & second ) const {
		int & data = * ( ( int * ) userData );

		data = 2;
		std::cout << first << " " << second << std::endl;
	}

	void Visit ( void * userData, const Tmp3 & first, const Tmp3 & second ) const {
		int & data = * ( ( int * ) userData );

		data = 3;
		std::cout << first << " " << second << std::endl;
	}

};

class TmpPromotingVisitor : public VisitableTmpBase::const_promoting_visitor_type {
	void Visit ( void * userData, const Tmp1 & first, const Tmp1 & second ) const {
		std::string & data = * ( ( std::string * ) userData );

		data = std::to_string ( first.getData ( ) ) + " " + std::to_string ( second.getData ( ) );
		std::cout << first << " " << second << std::endl;
	}

	void Visit ( void * userData, const Tmp2 & first, const Tmp2 & second ) const {
		std::string & data = * ( ( std::string * ) userData );

		data = std::to_string ( first.getData ( ) ) + " " + std::to_string ( second.getData ( ) );
		std::cout << first << " " << second << std::endl;
	}

	void Visit ( void * userData, const Tmp3 & first, const Tmp3 & second ) const {
		std::string & data = * ( ( std::string * ) userData );

		data = first.getData ( ) + " " + second.getData ( );
		std::cout << first << " " << second << std::endl;
	}

};

} /* namespace visitor */

void VisitorTest::testVisitor ( ) {
	visitor::TmpVisitor visitor;

	visitor::Tmp1 tmp1 ( 2 );

	int a = 0;

	tmp1.Accept ( ( void * ) & a, visitor );

	CPPUNIT_ASSERT ( a == 1 );
}

void VisitorTest::testSameVisitor ( ) {
	visitor::TmpSameVisitor visitor;

	visitor::Tmp2 tmpA ( 2 );
	visitor::Tmp2 tmpB ( 3 );

	int a = 0;

	Accept ( ( void * ) & a, ( visitor::TmpBase & ) tmpA, ( visitor::TmpBase & ) tmpB, visitor );

	CPPUNIT_ASSERT ( a == 2 );
}

void VisitorTest::testPromoteVisitor ( ) {
	visitor::TmpPromotingVisitor visitor;

	visitor::Tmp1 tmp1 ( 1 );
	visitor::Tmp3 tmp3 ( "3" );

	std::string a = "";

	Accept ( ( void * ) & a, ( visitor::TmpBase & ) tmp1, ( visitor::TmpBase & ) tmp3, visitor );

	CPPUNIT_ASSERT ( a == "1 3" );

	std::string b = "";
	Accept ( ( void * ) & b, ( visitor::TmpBase & ) tmp3, ( visitor::TmpBase & ) tmp1, visitor );

	CPPUNIT_ASSERT ( b == "3 1" );
}
