/*
 * Author: Radovan Cerveny
 */

#ifndef MEASUREMENT_FRAME_HPP_
#define MEASUREMENT_FRAME_HPP_

#include <chrono>
#include <sstream>
#include <deque>
#include "sax/Token.h"
#include "MeasurementTypes.hpp"

namespace measurements {

struct MeasurementFrame;

struct CounterHint {
	using value_type = long long int;
	enum class Type {
		ADD, SUB, INC, DEC
	};

	measurements::stealth_string name;
	Type						 type;
	value_type					 value;
};

struct CounterDataFrame {
	measurements::stealth_map < measurements::stealth_string, CounterHint::value_type > counters;
	measurements::stealth_map < measurements::stealth_string, CounterHint::value_type > real_counters;

	static void init ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void update ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void hint ( unsigned, measurements::stealth_vector < MeasurementFrame > &, CounterHint );
	static void xml_tokens ( std::deque < sax::Token > &, unsigned, const measurements::stealth_vector < MeasurementFrame > & );
};

std::ostream & operator <<( std::ostream &, const CounterDataFrame & );

struct MemoryHint {
	enum class Type {
		NEW, DELETE
	};

	Type   type;
	size_t size;
};

struct MemoryDataFrame {
	size_t start_heap_usage;
	size_t end_heap_usage;
	size_t high_watermark;

	size_t current_heap_usage;

	static void init ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void update ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void hint ( unsigned, measurements::stealth_vector < MeasurementFrame > &, MemoryHint );
	static void xml_tokens ( std::deque < sax::Token > &, unsigned, const measurements::stealth_vector < MeasurementFrame > & );
};

std::ostream & operator <<( std::ostream &, const MemoryDataFrame & );

struct TimeDataFrame {
	std::chrono::time_point < std::chrono::high_resolution_clock > start;

	std::chrono::microseconds duration;
	std::chrono::microseconds real_duration;

	static void init ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void update ( unsigned, measurements::stealth_vector < MeasurementFrame > & );
	static void xml_tokens ( std::deque < sax::Token > &, unsigned, const measurements::stealth_vector < MeasurementFrame > & );
};

std::ostream & operator <<( std::ostream &, const std::chrono::microseconds & );
std::ostream & operator <<( std::ostream &, const TimeDataFrame & );

struct MeasurementFrame {
	measurements::stealth_string name;
	measurements::Type			 type;
	unsigned					 parent_idx;

	measurements::stealth_vector < unsigned > sub_idxs;

	TimeDataFrame	 time;
	MemoryDataFrame	 memory;
	CounterDataFrame counter;

	MeasurementFrame ( measurements::stealth_string, measurements::Type, unsigned );
	static void xml_tokens ( std::deque < sax::Token > &, unsigned, const measurements::stealth_vector < MeasurementFrame > & );
};

std::ostream & operator <<( std::ostream &, const MeasurementFrame & );
}

#endif /* MEASUREMENT_FRAME_HPP_ */
