/*
 * Author: Radovan Cerveny
 */

#ifndef MEASUREMENT_TYPES_HPP_
#define MEASUREMENT_TYPES_HPP_

#include <memory>
#include <vector>
#include <map>
#include "MeasurementNew.hpp"

namespace measurements {

enum class Type : unsigned {
	ROOT = 1, OVERALL = 2, INIT = 4, FINALIZE = 8, MAIN = 16, AUXILIARY = 32
};

std::string to_string ( Type );
std::ostream & operator <<( std::ostream &, Type );

template < typename T, typename ... Args >
std::unique_ptr < T > make_unique ( Args && ... args ) {
	return std::unique_ptr < T > ( new T ( std::forward < Args > ( args ) ... ) );
}

template < typename T >
class stealth_allocator : public std::allocator < T > {
public:
	using pointer = T *;
	using size_type = size_t;

	pointer allocate ( size_type n ) {
		return static_cast < pointer > ( operator new( n * sizeof ( T ), false ) );
	}

	void deallocate ( pointer ptr, size_type ) {
		operator delete( ptr, false );
	}

	template < typename U >
	struct rebind { using other = stealth_allocator < U >; };

};

using stealth_string = std::basic_string < char, std::char_traits < char >, stealth_allocator < char > >;

std::string to_string ( const stealth_string & );
std::string to_string ( stealth_string & );

template < typename T >
using stealth_vector = std::vector < T, stealth_allocator < T > >;

template < typename T >
std::ostream & operator <<( std::ostream & os, const stealth_vector < T > & vec ) {
	os << "[";

	bool first = true;

	for ( const T & elem : vec ) {
		if ( !first ) os << ", ";

		first = false;
		os << elem;
	}

	return os << "]";
}

template < class Key, class T, class Compare = std::less < Key > >
using stealth_map = std::map < Key, T, Compare, stealth_allocator < std::pair < const Key, T > > >;
}

#endif /* MEASUREMENT_TYPES_HPP_ */
