/*
 * Author: Radovan Cerveny
 */

#include "MeasurementFrames.hpp"

namespace measurements {

void MemoryDataFrame::init ( unsigned frame_idx, measurements::stealth_vector < MeasurementFrame > & frames ) {
	MeasurementFrame & current_frame = frames[frame_idx];
	MeasurementFrame & parent_frame	 = frames[current_frame.parent_idx];

	if ( parent_frame.type != measurements::Type::ROOT ) {
		current_frame.memory.start_heap_usage = parent_frame.memory.current_heap_usage;
		current_frame.memory.current_heap_usage = parent_frame.memory.current_heap_usage;
		current_frame.memory.high_watermark = parent_frame.memory.current_heap_usage;
	}
}

void MemoryDataFrame::update ( unsigned frame_idx, measurements::stealth_vector < MeasurementFrame > & frames ) {
	MeasurementFrame & current_frame = frames[frame_idx];
	MeasurementFrame & parent_frame	 = frames[current_frame.parent_idx];

	current_frame.memory.end_heap_usage = current_frame.memory.current_heap_usage;

	 // if we end up with more memory from children, adjust watermark
	if ( current_frame.memory.current_heap_usage > current_frame.memory.high_watermark )
		current_frame.memory.high_watermark = current_frame.memory.current_heap_usage;

	if ( parent_frame.type != measurements::Type::ROOT )
		parent_frame.memory.current_heap_usage = current_frame.memory.current_heap_usage;
}

void MemoryDataFrame::hint ( unsigned frame_idx, measurements::stealth_vector < MeasurementFrame > & frames, MemoryHint hint ) {

	MeasurementFrame & current_frame = frames[frame_idx];

	switch ( hint.type ) {
	case MemoryHint::Type::NEW:
		current_frame.memory.current_heap_usage += hint.size;

		if ( current_frame.memory.current_heap_usage > current_frame.memory.high_watermark )
			current_frame.memory.high_watermark = current_frame.memory.current_heap_usage;

		break;

	case MemoryHint::Type::DELETE:
		current_frame.memory.current_heap_usage -= hint.size;
		break;
	}
}

void MemoryDataFrame::xml_tokens ( std::deque < sax::Token > & tokens, unsigned idx, const measurements::stealth_vector < MeasurementFrame > & frames ) {
	const MemoryDataFrame & frame = frames[idx].memory;

	tokens.emplace_back ( "MemoryData", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( "StartHeapUsage", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( std::to_string ( frame.start_heap_usage ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( "StartHeapUsage", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "EndHeapUsage", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( std::to_string ( frame.end_heap_usage ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( "EndHeapUsage", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "HighWatermark", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( std::to_string ( frame.high_watermark ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( "HighWatermark", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "MemoryData", sax::Token::TokenType::END_ELEMENT );
}

std::ostream & operator <<( std::ostream & os, const MemoryDataFrame & mdf ) {
	os << mdf.start_heap_usage << "shu, " << mdf.end_heap_usage << "ehu, " << mdf.high_watermark << "hw";
	return os;
}

}
