/*
 * vector.hpp
 *
 * Created on: Feb 28, 2014
 * Author: Jan Travnicek
 */

#ifndef __VECTOR_HPP_
#define __VECTOR_HPP_

namespace std {

template <class Iterator>
class ptr_vector_iterator {
	Iterator current;

public:
	typedef Iterator iterator_type;
	typedef typename iterator_traits<Iterator>::iterator_category iterator_category;
	typedef typename iterator_traits<Iterator>::value_type        value_type;
	typedef Iterator pointer;
	typedef typename std::remove_pointer<value_type>::type& reference;

	explicit ptr_vector_iterator() {
	}

	explicit ptr_vector_iterator (Iterator it) : current(it) {
	}

	ptr_vector_iterator (const ptr_vector_iterator<Iterator>& it) : current(it.current) {
	}

	ptr_vector_iterator& operator= (const ptr_vector_iterator<Iterator>& it) {
		current = it.current;
	}

	iterator_type base() const {
		return current;
	}

	pointer operator->() const {
		return *current;
	}

	reference operator*() const {
		return **current;
	}

	ptr_vector_iterator& operator++() {
		++current;
		return *this;
	}

	ptr_vector_iterator& operator--() {
		--current;
		return *this;
	}

	ptr_vector_iterator operator++(int) {
		ptr_vector_iterator temp = *this;
		++current;
		return temp;
	}

	ptr_vector_iterator operator--(int) {
		ptr_vector_iterator temp = *this;
		--current;
		return temp;
	}

	bool operator== (const ptr_vector_iterator<Iterator>& other) {
		return this->current == other.current;
	}

	bool operator!= (const ptr_vector_iterator<Iterator>& other) {
		return ! ( *this == other );
	}

};

template<class Iterator>
ptr_vector_iterator<Iterator> make_ptr_vector_iterator (Iterator it) {
	return ptr_vector_iterator<Iterator>(it);
}


template<class T>
class ptr_vector {
	vector<T*> data;

public:
	ptr_vector() {}

	ptr_vector(const ptr_vector<T>& other) {
		data.reserve(other.size());
		for(T* ptr : other.data) {
			data.push_back(std::clone(ptr));
		}
	}

	ptr_vector(ptr_vector<T>&& other) noexcept {
		std::swap(this->data, other.data);
	}

	ptr_vector<T>& operator=(const ptr_vector<T>& other) {
		if(this == *other) return *this;

		*this = ptr_vector(other);

		return *this;
	}

	~ptr_vector() noexcept {
		for(T* elem : data) {
			delete elem;
		}
	}

	ptr_vector<T>& operator=(ptr_vector<T>&& other) noexcept {
		std::swap(this->data, other.data);
		return *this;
	}

	T& operator[]( int n ) {
		return *data[n];
	}

	const T& operator[]( int n ) const {
		return *data[n];
	}

	void push_back(T* elem) {
		data.push_back(elem);
	}

	unsigned size() const {
		return data.size();
	}

	ptr_vector_iterator<typename std::vector<T*>::iterator> begin() {
		return make_ptr_vector_iterator(data.begin());
	}

	ptr_vector_iterator<typename std::vector<T*>::iterator> end() {
		return make_ptr_vector_iterator(data.end());
	}

	ptr_vector_iterator<typename std::vector<T*>::const_iterator> begin() const {
		return make_ptr_vector_iterator(data.cbegin());
	}

	ptr_vector_iterator<typename std::vector<T*>::const_iterator> end() const {
		return make_ptr_vector_iterator(data.cend());
	}

	ptr_vector_iterator<typename std::vector<T*>::const_iterator> cbegin() {
		return make_ptr_vector_iterator(data.cbegin());
	}

	ptr_vector_iterator<typename std::vector<T*>::const_iterator> cend() {
		return make_ptr_vector_iterator(data.cend());
	}
};

template< class T >
std::ostream& operator<<(std::ostream& out, const std::vector<T>& vector) {
	out << "[";

	bool first = true;
	for(const T& item : vector) {
		if(!first) out << ", ";
		first = false;
		out << item;
	}

	out << "]";
	return out;
}

template< class T >
std::ostream& operator<<(std::ostream& out, const std::ptr_vector<T>& vector) {
	out << "[";

	bool first = true;
	for(const T& item : vector) {
		if(!first) out << ", ";
		first = false;
		out << item;
	}

	out << "]";
	return out;
}

template<class T>
struct compare<vector<T>> {
	int operator()(const vector<T>& first, const vector<T>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<T> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); ++iterF, ++iterS) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

template<class T>
struct compare<ptr_vector<T>> {
	int operator()(const ptr_vector<T>& first, const ptr_vector<T>& second) const {
		if(first.size() < second.size()) return -1;
		if(first.size() > second.size()) return 1;

		compare<T> comp;
		for(auto iterF = first.begin(), iterS = second.begin(); iterF != first.end(); ++iterF, ++iterS) {
			int res = comp(*iterF, *iterS);
			if(res != 0) return res;
		}
		return 0;
	}
};

} /* namespace std */

#endif /* __VECTOR_HPP_ */

