/*
 * astringology.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <common/GlobalData.h>
#include <measure>
#include <vector>

#include <factory/XmlDataFactory.hpp>
#include <exception/AlibException.h>
#include <string/String.h>
#include <automaton/Automaton.h>
#include <container/Container.h>
#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include <stringology/exact/ExactFactorMatch.h>
#include <stringology/exact/BoyerMooreHorspool.h>
#include <stringology/exact/ReversedBoyerMooreHorspool.h>
#include <stringology/exact/DeadZoneUsingBadCharacterShift.h>
#include <stringology/exact/ExactMatchingAutomaton.h>
#include <stringology/exact/ExactFactorAutomaton.h>
#include <stringology/exact/ExactSubsequenceAutomaton.h>
#include <stringology/exact/ExactNondeterministicSubsequenceAutomaton.h>
#include <stringology/exact/ExactMultiNondeterministicSubsequenceAutomaton.h>
#include <stringology/exact/BorderArray.h>
#include <stringology/indexing/SuffixTrie.h>

int main ( int argc, char * argv[] ) {
	try {
		TCLAP::CmdLine cmd ( "Stringology algorithm access binary", ' ', "0.01" );

		std::vector < std::string > allowed;
		allowed.push_back ( "exactMatchingAutomaton" );
		allowed.push_back ( "exactFactorAutomaton" );
		allowed.push_back ( "exactSubsequenceAutomaton" );
		allowed.push_back ( "exactNondeterministicSubsequenceAutomaton" );
		allowed.push_back ( "exactMultiNondeterministicSubsequenceAutomaton" );
		allowed.push_back ( "exactFactorMatch" );
		allowed.push_back ( "boyerMooreHorspool" );
		allowed.push_back ( "reversedBoyerMooreHorspool" );
		allowed.push_back ( "deadZoneUsingBadCharacterShift" );
		allowed.push_back ( "borderArray" );
		allowed.push_back ( "suffixTrie" );
		TCLAP::ValuesConstraint < std::string > allowedVals ( allowed );

		TCLAP::ValueArg < std::string > algorithm ( "a", "algorithm", "Execute algorithm", false, "exactFactorMatch", & allowedVals );
		cmd.add ( algorithm );

		TCLAP::MultiArg < std::string > subject ( "s", "subject", "Subject string from file", false, "file" );
		cmd.add ( subject );

		TCLAP::MultiArg < std::string > pattern ( "p", "pattern", "Pattern string from file", false, "file" );
		cmd.add ( pattern );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		int needPattern = 0;
		int needSubject = 0;

		if ( algorithm.getValue ( ) == "exactFactorMatch" ) {
			needPattern = needSubject = 1;
		} else if ( algorithm.getValue ( ) == "boyerMooreHorspool" ) {
			needPattern = needSubject = 1;
		} else if ( algorithm.getValue ( ) == "reversedBoyerMooreHorspool" ) {
			needPattern = needSubject = 1;
		} else if ( algorithm.getValue ( ) == "deadZoneUsingBadCharacterShift" ) {
			needPattern = needSubject = 1;
		} else if ( algorithm.getValue ( ) == "exactMatchingAutomaton" ) {
			needPattern = 1;
		} else if ( algorithm.getValue ( ) == "exactFactorAutomaton" ) {
			needSubject = 1;
		} else if ( algorithm.getValue ( ) == "exactSubsequenceAutomaton" ) {
			needSubject = 1;
		} else if ( algorithm.getValue ( ) == "exactNondeterministicSubsequenceAutomaton" ) {
			needSubject = 1;
		} else if ( algorithm.getValue ( ) == "exactMultiNondeterministicSubsequenceAutomaton" ) {
			needSubject = 2;
		} else if ( algorithm.getValue ( ) == "borderArray" ) {
			needSubject = 1;
		} else {
		}

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		std::deque < std::deque < sax::Token > > subjectTokens;

		if ( subject.isSet ( ) ) {
			for ( const std::string & fileName : subject.getValue ( ) ) {
				std::deque < sax::Token > tmp;

				if ( fileName == "-" )
					sax::SaxParseInterface::parseStdin ( tmp );
				else
					sax::SaxParseInterface::parseFile ( fileName, tmp );

				subjectTokens.emplace_back ( std::move ( tmp ) );
			}
		} else if ( needSubject ) {
			std::deque < sax::Token > tmp;
			sax::SaxParseInterface::parseStdin ( tmp );
			subjectTokens.emplace_back ( std::move ( tmp ) );
		}

		std::deque < std::deque < sax::Token > > patternTokens;

		if ( pattern.isSet ( ) ) {
			for ( const std::string & fileName : pattern.getValue ( ) ) {
				std::deque < sax::Token > tmp;

				if ( fileName == "-" )
					sax::SaxParseInterface::parseStdin ( tmp );
				else
					sax::SaxParseInterface::parseFile ( fileName, tmp );

				patternTokens.emplace_back ( std::move ( tmp ) );
			}
		} else if ( needPattern ) {
			std::deque < sax::Token > tmp;
			sax::SaxParseInterface::parseStdin ( tmp );
			patternTokens.emplace_back ( std::move ( tmp ) );
		}

		if ( algorithm.getValue ( ) == "exactFactorMatch" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );
			string::String pattern = alib::XmlDataFactory::fromTokens < string::String > ( patternTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::ExactFactorMatch::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "boyerMooreHorspool" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );
			string::String pattern = alib::XmlDataFactory::fromTokens < string::String > ( patternTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::BoyerMooreHorspool::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "reversedBoyerMooreHorspool" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );
			string::String pattern = alib::XmlDataFactory::fromTokens < string::String > ( patternTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::ReversedBoyerMooreHorspool::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "deadZoneUsingBadCharacterShift" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );
			string::String pattern = alib::XmlDataFactory::fromTokens < string::String > ( patternTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::DeadZoneUsingBadCharacterShift::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "exactMatchingAutomaton" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens < string::String > ( patternTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::exact::ExactMatchingAutomaton::construct ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactFactorAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::exact::ExactFactorAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactSubsequenceAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::exact::ExactSubsequenceAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactNondeterministicSubsequenceAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::exact::ExactNondeterministicSubsequenceAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactMultiNondeterministicSubsequenceAutomaton" ) {
			std::set < string::String > subjects;

			for ( std::deque < sax::Token > & tokens : subjectTokens )
				subjects.insert ( alib::XmlDataFactory::fromTokens < string::String > ( tokens ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::exact::ExactMultiNondeterministicSubsequenceAutomaton::construct ( subjects );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "borderArray" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::vector < unsigned > borderArray = stringology::exact::BorderArray::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( borderArray );
		} else if ( algorithm.getValue ( ) == "suffixTrie" ) {
			string::String subject = alib::XmlDataFactory::fromTokens < string::String > ( subjectTokens.front ( ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::SuffixTrieFinalMark suffixTrie = stringology::indexing::SuffixTrie::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( suffixTrie );
		} else {
			throw exception::AlibException ( "Invalid algorithm" );
		}

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) std::clog << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::AlibException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
